import logging
import os
import sys
import time
import warnings
from asyncio import iscoroutinefunction
from functools import wraps
from itertools import combinations
from threading import local
from typing import Any, Callable, Optional, TextIO, Union
from urllib.parse import quote, unquote, urlparse

from neo4j import (
    DEFAULT_DATABASE,
    AsyncDriver,
    AsyncGraphDatabase,
    AsyncResult,
    AsyncSession,
    AsyncTransaction,
    basic_auth,
)
from neo4j.api import Bookmarks
from neo4j.exceptions import ClientError, ServiceUnavailable, SessionExpired
from neo4j.graph import Node, Path, Relationship

from neomodel import config
from neomodel._async_compat.util import AsyncUtil
from neomodel.async_.property_manager import AsyncPropertyManager
from neomodel.exceptions import (
    ConstraintValidationFailed,
    DoesNotExist,
    FeatureNotSupported,
    NodeClassAlreadyDefined,
    NodeClassNotDefined,
    RelationshipClassNotDefined,
    UniqueProperty,
)
from neomodel.hooks import hooks
from neomodel.properties import FulltextIndex, Property, VectorIndex
from neomodel.util import (
    _UnsavedNode,
    classproperty,
    deprecated,
    version_tag_to_integer,
)

logger = logging.getLogger(__name__)

RULE_ALREADY_EXISTS = "Neo.ClientError.Schema.EquivalentSchemaRuleAlreadyExists"
INDEX_ALREADY_EXISTS = "Neo.ClientError.Schema.IndexAlreadyExists"
CONSTRAINT_ALREADY_EXISTS = "Neo.ClientError.Schema.ConstraintAlreadyExists"
STREAMING_WARNING = "streaming is not supported by bolt, please remove the kwarg"
NOT_COROUTINE_ERROR = "The decorated function must be a coroutine"


# make sure the connection url has been set prior to executing the wrapped function
def ensure_connection(func: Callable) -> Callable:
    """Decorator that ensures a connection is established before executing the decorated function.

    Args:
        func (callable): The function to be decorated.

    Returns:
        callable: The decorated function.

    """

    async def wrapper(self: Any, *args: Any, **kwargs: Any) -> Callable:
        # Sort out where to find url
        if hasattr(self, "db"):
            _db = self.db
        else:
            _db = self

        if not _db.driver:
            if hasattr(config, "DATABASE_URL") and config.DATABASE_URL:
                await _db.set_connection(url=config.DATABASE_URL)
            elif hasattr(config, "DRIVER") and config.DRIVER:
                await _db.set_connection(driver=config.DRIVER)

        return await func(self, *args, **kwargs)

    return wrapper


class AsyncDatabase(local):
    """
    A singleton object via which all operations from neomodel to the Neo4j backend are handled with.
    """

    _NODE_CLASS_REGISTRY: dict[frozenset, Any] = {}
    _DB_SPECIFIC_CLASS_REGISTRY: dict[str, dict[frozenset, Any]] = {}

    def __init__(self) -> None:
        self._active_transaction: Optional[AsyncTransaction] = None
        self.url: Optional[str] = None
        self.driver: Optional[AsyncDriver] = None
        self._session: Optional[AsyncSession] = None
        self._pid: Optional[int] = None
        self._database_name: Optional[str] = DEFAULT_DATABASE
        self._database_version: Optional[str] = None
        self._database_edition: Optional[str] = None
        self.impersonated_user: Optional[str] = None
        self._parallel_runtime: Optional[bool] = False

    async def set_connection(
        self, url: Optional[str] = None, driver: Optional[AsyncDriver] = None
    ) -> None:
        """
        Sets the connection up and relevant internal. This can be done using a Neo4j URL or a driver instance.

        Args:
            url (str): Optionally, Neo4j URL in the form protocol://username:password@hostname:port/dbname.
            When provided, a Neo4j driver instance will be created by neomodel.

            driver (neo4j.Driver): Optionally, a pre-created driver instance.
            When provided, neomodel will not create a driver instance but use this one instead.
        """
        if driver:
            self.driver = driver
            if hasattr(config, "DATABASE_NAME") and config.DATABASE_NAME:
                self._database_name = config.DATABASE_NAME
        elif url:
            self._parse_driver_from_url(url=url)

        self._pid = os.getpid()
        self._active_transaction = None
        # Set to default database if it hasn't been set before
        if self._database_name is None:
            self._database_name = DEFAULT_DATABASE

        # Getting the information about the database version requires a connection to the database
        self._database_version = None
        self._database_edition = None
        await self._update_database_version()

    def _parse_driver_from_url(self, url: str) -> None:
        """Parse the driver information from the given URL and initialize the driver.

        Args:
            url (str): The URL to parse.

        Raises:
            ValueError: If the URL format is not as expected.

        Returns:
            None - Sets the driver and database_name as class properties
        """
        p_start = url.replace(":", "", 1).find(":") + 2
        p_end = url.rfind("@")
        password = url[p_start:p_end]
        url = url.replace(password, quote(password))
        parsed_url = urlparse(url)

        valid_schemas = [
            "bolt",
            "bolt+s",
            "bolt+ssc",
            "bolt+routing",
            "neo4j",
            "neo4j+s",
            "neo4j+ssc",
        ]

        if parsed_url.netloc.find("@") > -1 and parsed_url.scheme in valid_schemas:
            credentials, hostname = parsed_url.netloc.rsplit("@", 1)
            username, password = credentials.split(":")
            password = unquote(password)
            database_name = parsed_url.path.strip("/")
        else:
            raise ValueError(
                f"Expecting url format: bolt://user:password@localhost:7687 got {url}"
            )

        options = {
            "auth": basic_auth(username, password),
            "connection_acquisition_timeout": config.CONNECTION_ACQUISITION_TIMEOUT,
            "connection_timeout": config.CONNECTION_TIMEOUT,
            "keep_alive": config.KEEP_ALIVE,
            "max_connection_lifetime": config.MAX_CONNECTION_LIFETIME,
            "max_connection_pool_size": config.MAX_CONNECTION_POOL_SIZE,
            "max_transaction_retry_time": config.MAX_TRANSACTION_RETRY_TIME,
            "resolver": config.RESOLVER,
            "user_agent": config.USER_AGENT,
        }

        if "+s" not in parsed_url.scheme:
            options["encrypted"] = config.ENCRYPTED
            options["trusted_certificates"] = config.TRUSTED_CERTIFICATES

        # Ignore the type error because the workaround would be duplicating code
        self.driver = AsyncGraphDatabase.driver(
            parsed_url.scheme + "://" + hostname, **options  # type: ignore[arg-type]
        )
        self.url = url
        # The database name can be provided through the url or the config
        if database_name == "":
            if hasattr(config, "DATABASE_NAME") and config.DATABASE_NAME:
                self._database_name = config.DATABASE_NAME
        else:
            self._database_name = database_name

    async def close_connection(self) -> None:
        """
        Closes the currently open driver.
        The driver should always be closed at the end of the application's lifecyle.
        """
        self._database_version = None
        self._database_edition = None
        self._database_name = None
        if self.driver is not None:
            await self.driver.close()
            self.driver = None

    @property
    async def database_version(self) -> Optional[str]:
        if self._database_version is None:
            await self._update_database_version()

        return self._database_version

    @property
    async def database_edition(self) -> Optional[str]:
        if self._database_edition is None:
            await self._update_database_version()

        return self._database_edition

    @property
    def transaction(self) -> "AsyncTransactionProxy":
        """
        Returns the current transaction object
        """
        return AsyncTransactionProxy(self)

    @property
    def write_transaction(self) -> "AsyncTransactionProxy":
        return AsyncTransactionProxy(self, access_mode="WRITE")

    @property
    def read_transaction(self) -> "AsyncTransactionProxy":
        return AsyncTransactionProxy(self, access_mode="READ")

    @property
    def parallel_read_transaction(self) -> "AsyncTransactionProxy":
        return AsyncTransactionProxy(self, access_mode="READ", parallel_runtime=True)

    async def impersonate(self, user: str) -> "ImpersonationHandler":
        """All queries executed within this context manager will be executed as impersonated user

        Args:
            user (str): User to impersonate

        Returns:
            ImpersonationHandler: Context manager to set/unset the user to impersonate
        """
        db_edition = await self.database_edition
        if db_edition != "enterprise":
            raise FeatureNotSupported(
                "Impersonation is only available in Neo4j Enterprise edition"
            )
        return ImpersonationHandler(self, impersonated_user=user)

    @ensure_connection
    async def begin(self, access_mode: str = "WRITE", **parameters: Any) -> None:
        """
        Begins a new transaction. Raises SystemError if a transaction is already active.
        """
        if (
            hasattr(self, "_active_transaction")
            and self._active_transaction is not None
        ):
            raise SystemError("Transaction in progress")

        assert self.driver is not None, "Driver has not been created"

        self._session = self.driver.session(
            default_access_mode=access_mode,
            database=self._database_name,
            impersonated_user=self.impersonated_user,
            **parameters,
        )

        assert self._session is not None, "Session has not been created"
        self._active_transaction = await self._session.begin_transaction()

    @ensure_connection
    async def commit(self) -> Bookmarks:
        """
        Commits the current transaction and closes its session

        :return: last_bookmarks
        """
        try:
            assert self._active_transaction is not None, "No transaction in progress"
            await self._active_transaction.commit()

            assert self._session is not None, "No session open"
            last_bookmarks: Bookmarks = await self._session.last_bookmarks()
        finally:
            # In case something went wrong during
            # committing changes to the database
            # we have to close an active transaction and session.
            assert self._active_transaction is not None, "No transaction in progress"
            await self._active_transaction.close()

            assert self._session is not None, "No session open"
            await self._session.close()

            self._active_transaction = None
            self._session = None

        return last_bookmarks

    @ensure_connection
    async def rollback(self) -> None:
        """
        Rolls back the current transaction and closes its session
        """
        try:
            assert self._active_transaction is not None, "No transaction in progress"
            await self._active_transaction.rollback()
        finally:
            # In case when something went wrong during changes rollback,
            # we have to close an active transaction and session
            assert self._active_transaction is not None, "No transaction in progress"
            await self._active_transaction.close()

            assert self._session is not None, "No session open"
            await self._session.close()

            self._active_transaction = None
            self._session = None

    async def _update_database_version(self) -> None:
        """
        Updates the database server information when it is required
        """
        try:
            results = await self.cypher_query(
                "CALL dbms.components() yield versions, edition return versions[0], edition"
            )
            self._database_version = results[0][0][0]
            self._database_edition = results[0][0][1]
        except ServiceUnavailable:
            # The database server is not running yet
            pass

    def _object_resolution(self, object_to_resolve: Any) -> Any:
        """
        Performs in place automatic object resolution on a result
        returned by cypher_query.

        The function operates recursively in order to be able to resolve Nodes
        within nested list structures and Path objects. Not meant to be called
        directly, used primarily by _result_resolution.

        :param object_to_resolve: A result as returned by cypher_query.
        :type Any:

        :return: An instantiated object.
        """
        # Below is the original comment that came with the code extracted in
        # this method. It is not very clear but I decided to keep it just in
        # case
        #
        #
        # For some reason, while the type of `a_result_attribute[1]`
        # as reported by the neo4j driver is `Node` for Node-type data
        # retrieved from the database.
        # When the retrieved data are Relationship-Type,
        # the returned type is `abc.[REL_LABEL]` which is however
        # a descendant of Relationship.
        # Consequently, the type checking was changed for both
        # Node, Relationship objects
        if isinstance(object_to_resolve, Node):
            _labels = frozenset(object_to_resolve.labels)
            if _labels in self._NODE_CLASS_REGISTRY:
                return self._NODE_CLASS_REGISTRY[_labels].inflate(object_to_resolve)
            elif (
                self._database_name is not None
                and self._database_name in self._DB_SPECIFIC_CLASS_REGISTRY
                and _labels in self._DB_SPECIFIC_CLASS_REGISTRY[self._database_name]
            ):
                return self._DB_SPECIFIC_CLASS_REGISTRY[self._database_name][
                    _labels
                ].inflate(object_to_resolve)
            else:
                raise NodeClassNotDefined(
                    object_to_resolve,
                    self._NODE_CLASS_REGISTRY,
                    self._DB_SPECIFIC_CLASS_REGISTRY,
                )

        if isinstance(object_to_resolve, Relationship):
            rel_type = frozenset([object_to_resolve.type])
            if rel_type in self._NODE_CLASS_REGISTRY:
                return self._NODE_CLASS_REGISTRY[rel_type].inflate(object_to_resolve)
            elif (
                self._database_name is not None
                and self._database_name in self._DB_SPECIFIC_CLASS_REGISTRY
                and rel_type in self._DB_SPECIFIC_CLASS_REGISTRY[self._database_name]
            ):
                return self._DB_SPECIFIC_CLASS_REGISTRY[self._database_name][
                    rel_type
                ].inflate(object_to_resolve)
            else:
                raise RelationshipClassNotDefined(
                    object_to_resolve,
                    self._NODE_CLASS_REGISTRY,
                    self._DB_SPECIFIC_CLASS_REGISTRY,
                )

        if isinstance(object_to_resolve, Path):
            from neomodel.async_.path import AsyncNeomodelPath

            return AsyncNeomodelPath(object_to_resolve)

        if isinstance(object_to_resolve, list):
            return self._result_resolution([object_to_resolve])

        return object_to_resolve

    def _result_resolution(self, result_list: list) -> list:
        """
        Performs in place automatic object resolution on a set of results
        returned by cypher_query.

        The function operates recursively in order to be able to resolve Nodes
        within nested list structures. Not meant to be called directly,
        used primarily by cypher_query.

        :param result_list: A list of results as returned by cypher_query.
        :type list:

        :return: A list of instantiated objects.
        """

        # Object resolution occurs in-place
        for a_result_item in enumerate(result_list):
            for a_result_attribute in enumerate(a_result_item[1]):
                # Primitive types should remain primitive types,
                # Nodes to be resolved to native objects
                resolved_object = a_result_attribute[1]

                resolved_object = self._object_resolution(resolved_object)

                result_list[a_result_item[0]][a_result_attribute[0]] = resolved_object

        return result_list

    @ensure_connection
    async def cypher_query(
        self,
        query: str,
        params: Optional[dict[str, Any]] = None,
        handle_unique: bool = True,
        retry_on_session_expire: bool = False,
        resolve_objects: bool = False,
    ) -> tuple[Optional[list], Optional[tuple[str, ...]]]:
        """
        Runs a query on the database and returns a list of results and their headers.

        :param query: A CYPHER query
        :type: str
        :param params: Dictionary of parameters
        :type: dict
        :param handle_unique: Whether or not to raise UniqueProperty exception on Cypher's ConstraintValidation errors
        :type: bool
        :param retry_on_session_expire: Whether or not to attempt the same query again if the transaction has expired.
        If you use neomodel with your own driver, you must catch SessionExpired exceptions yourself and retry with a new driver instance.
        :type: bool
        :param resolve_objects: Whether to attempt to resolve the returned nodes to data model objects automatically
        :type: bool

        :return: A tuple containing a list of results and a tuple of headers.
        """
        if params is None:
            params = {}
        if self._active_transaction:
            # Use current transaction if a transaction is currently active
            results, meta = await self._run_cypher_query(
                self._active_transaction,
                query,
                params,
                handle_unique,
                retry_on_session_expire,
                resolve_objects,
            )
        else:
            # Otherwise create a new session in a with to dispose of it after it has been run
            if self.driver:
                async with self.driver.session(
                    database=self._database_name,
                    impersonated_user=self.impersonated_user,
                ) as session:
                    results, meta = await self._run_cypher_query(
                        session,
                        query,
                        params,
                        handle_unique,
                        retry_on_session_expire,
                        resolve_objects,
                    )
            else:
                raise ValueError("No driver has been set")

        return results, meta

    async def _run_cypher_query(
        self,
        session: Union[AsyncSession, AsyncTransaction],
        query: str,
        params: dict[str, Any],
        handle_unique: bool,
        retry_on_session_expire: bool,
        resolve_objects: bool,
    ) -> tuple[Optional[list], Optional[tuple[str, ...]]]:
        try:
            # Retrieve the data
            start = time.time()
            if self._parallel_runtime:
                query = "CYPHER runtime=parallel " + query
            response: AsyncResult = await session.run(query=query, parameters=params)
            results, meta = [list(r.values()) async for r in response], response.keys()
            end = time.time()

            if resolve_objects:
                # Do any automatic resolution required
                results = self._result_resolution(results)

        except ClientError as e:
            if e.code == "Neo.ClientError.Schema.ConstraintValidationFailed":
                if hasattr(e, "message") and e.message is not None:
                    if "already exists with label" in e.message and handle_unique:
                        raise UniqueProperty(e.message) from e
                    raise ConstraintValidationFailed(e.message) from e
                raise ConstraintValidationFailed(
                    "A constraint validation failed"
                ) from e

            exc_info = sys.exc_info()
            if exc_info[1] is not None and exc_info[2] is not None:
                raise exc_info[1].with_traceback(exc_info[2])
        except SessionExpired:
            if retry_on_session_expire:
                await self.set_connection(url=self.url)
                return await self.cypher_query(
                    query=query,
                    params=params,
                    handle_unique=handle_unique,
                    retry_on_session_expire=False,
                )
            raise

        tte = end - start
        if os.environ.get("NEOMODEL_CYPHER_DEBUG", False) and tte > float(
            os.environ.get("NEOMODEL_SLOW_QUERIES", 0)
        ):
            logger.debug(
                "query: "
                + query
                + "\nparams: "
                + repr(params)
                + f"\ntook: {tte:.2g}s\n"
            )

        return results, meta

    async def get_id_method(self) -> str:
        db_version = await self.database_version
        if db_version is None:
            raise RuntimeError(
                """
                Unable to perform this operation because the database server version is not known. 
                This might mean that the database server is offline.
                """
            )
        if db_version.startswith("4"):
            return "id"
        else:
            return "elementId"

    async def parse_element_id(self, element_id: Optional[str]) -> Union[str, int]:
        if element_id is None:
            raise ValueError(
                "Unable to parse element id, are you sure this element has been saved ?"
            )
        db_version = await self.database_version
        if db_version is None:
            raise RuntimeError(
                """
                Unable to perform this operation because the database server version is not known. 
                This might mean that the database server is offline.
                """
            )
        return int(element_id) if db_version.startswith("4") else element_id

    async def list_indexes(self, exclude_token_lookup: bool = False) -> list[dict]:
        """Returns all indexes existing in the database

        Arguments:
            exclude_token_lookup[bool]: Exclude automatically create token lookup indexes

        Returns:
            Sequence[dict]: List of dictionaries, each entry being an index definition
        """
        indexes, meta_indexes = await self.cypher_query("SHOW INDEXES")
        indexes_as_dict = [dict(zip(meta_indexes, row)) for row in indexes]

        if exclude_token_lookup:
            indexes_as_dict = [
                obj for obj in indexes_as_dict if obj["type"] != "LOOKUP"
            ]

        return indexes_as_dict

    async def list_constraints(self) -> list[dict]:
        """Returns all constraints existing in the database

        Returns:
            Sequence[dict]: List of dictionaries, each entry being a constraint definition
        """
        constraints, meta_constraints = await self.cypher_query("SHOW CONSTRAINTS")
        constraints_as_dict = [dict(zip(meta_constraints, row)) for row in constraints]

        return constraints_as_dict

    @ensure_connection
    async def version_is_higher_than(self, version_tag: str) -> bool:
        """Returns true if the database version is higher or equal to a given tag

        Args:
            version_tag (str): The version to compare against

        Returns:
            bool: True if the database version is higher or equal to the given version
        """
        db_version = await self.database_version
        if db_version is None:
            raise RuntimeError(
                """
                Unable to perform this operation because the database server version is not known. 
                This might mean that the database server is offline.
                """
            )
        return version_tag_to_integer(db_version) >= version_tag_to_integer(version_tag)

    @ensure_connection
    async def edition_is_enterprise(self) -> bool:
        """Returns true if the database edition is enterprise

        Returns:
            bool: True if the database edition is enterprise
        """
        edition = await self.database_edition
        if edition is None:
            raise RuntimeError(
                """
                Unable to perform this operation because the database server edition is not known. 
                This might mean that the database server is offline.
                """
            )
        return edition == "enterprise"

    @ensure_connection
    async def parallel_runtime_available(self) -> bool:
        """Returns true if the database supports parallel runtime

        Returns:
            bool: True if the database supports parallel runtime
        """
        return (
            await self.version_is_higher_than("5.13")
            and await self.edition_is_enterprise()
        )

    async def change_neo4j_password(self, user: str, new_password: str) -> None:
        await self.cypher_query(f"ALTER USER {user} SET PASSWORD '{new_password}'")

    async def clear_neo4j_database(
        self, clear_constraints: bool = False, clear_indexes: bool = False
    ) -> None:
        await self.cypher_query(
            """
            MATCH (a)
            CALL { WITH a DETACH DELETE a }
            IN TRANSACTIONS OF 5000 rows
        """
        )
        if clear_constraints:
            await drop_constraints()
        if clear_indexes:
            await drop_indexes()

    async def drop_constraints(
        self, quiet: bool = True, stdout: Optional[TextIO] = None
    ) -> None:
        """
        Discover and drop all constraints.

        :type: bool
        :return: None
        """
        if not stdout or stdout is None:
            stdout = sys.stdout

        results, meta = await self.cypher_query("SHOW CONSTRAINTS")

        results_as_dict = [dict(zip(meta, row)) for row in results]
        for constraint in results_as_dict:
            await self.cypher_query("DROP CONSTRAINT " + constraint["name"])
            if not quiet:
                stdout.write(
                    (
                        " - Dropping unique constraint and index"
                        f" on label {constraint['labelsOrTypes'][0]}"
                        f" with property {constraint['properties'][0]}.\n"
                    )
                )
        if not quiet:
            stdout.write("\n")

    async def drop_indexes(
        self, quiet: bool = True, stdout: Optional[TextIO] = None
    ) -> None:
        """
        Discover and drop all indexes, except the automatically created token lookup indexes.

        :type: bool
        :return: None
        """
        if not stdout or stdout is None:
            stdout = sys.stdout

        indexes = await self.list_indexes(exclude_token_lookup=True)
        for index in indexes:
            await self.cypher_query("DROP INDEX " + index["name"])
            if not quiet:
                stdout.write(
                    f' - Dropping index on labels {",".join(index["labelsOrTypes"])} with properties {",".join(index["properties"])}.\n'
                )
        if not quiet:
            stdout.write("\n")

    async def remove_all_labels(self, stdout: Optional[TextIO] = None) -> None:
        """
        Calls functions for dropping constraints and indexes.

        :param stdout: output stream
        :return: None
        """

        if not stdout:
            stdout = sys.stdout

        stdout.write("Dropping constraints...\n")
        await self.drop_constraints(quiet=False, stdout=stdout)

        stdout.write("Dropping indexes...\n")
        await self.drop_indexes(quiet=False, stdout=stdout)

    async def install_all_labels(self, stdout: Optional[TextIO] = None) -> None:
        """
        Discover all subclasses of StructuredNode in your application and execute install_labels on each.
        Note: code must be loaded (imported) in order for a class to be discovered.

        :param stdout: output stream
        :return: None
        """

        if not stdout or stdout is None:
            stdout = sys.stdout

        def subsub(cls: Any) -> list:  # recursively return all subclasses
            subclasses = cls.__subclasses__()
            if not subclasses:  # base case: no more subclasses
                return []
            return subclasses + [g for s in cls.__subclasses__() for g in subsub(s)]

        stdout.write("Setting up indexes and constraints...\n\n")

        i = 0
        for cls in subsub(AsyncStructuredNode):
            stdout.write(f"Found {cls.__module__}.{cls.__name__}\n")
            await install_labels(cls, quiet=False, stdout=stdout)
            i += 1

        if i:
            stdout.write("\n")

        stdout.write(f"Finished {i} classes.\n")

    async def install_labels(
        self, cls: Any, quiet: bool = True, stdout: Optional[TextIO] = None
    ) -> None:
        """
        Setup labels with indexes and constraints for a given class

        :param cls: StructuredNode class
        :type: class
        :param quiet: (default true) enable standard output
        :param stdout: stdout stream
        :type: bool
        :return: None
        """
        _stdout = stdout if stdout else sys.stdout

        if not hasattr(cls, "__label__"):
            if not quiet:
                _stdout.write(
                    f" ! Skipping class {cls.__module__}.{cls.__name__} is abstract\n"
                )
            return

        for name, property in cls.defined_properties(aliases=False, rels=False).items():
            await self._install_node(cls, name, property, quiet, _stdout)

        for _, relationship in cls.defined_properties(
            aliases=False, rels=True, properties=False
        ).items():
            await self._install_relationship(cls, relationship, quiet, _stdout)

    async def _create_node_index(
        self, target_cls: Any, property_name: str, stdout: TextIO, quiet: bool
    ) -> None:
        label = target_cls.__label__
        index_name = f"index_{label}_{property_name}"
        if not quiet:
            stdout.write(
                f" + Creating node index for {property_name} on label {label} for class {target_cls.__module__}.{target_cls.__name__}\n"
            )
        try:
            await self.cypher_query(
                f"CREATE INDEX {index_name} FOR (n:{label}) ON (n.{property_name}); "
            )
        except ClientError as e:
            if e.code in (
                RULE_ALREADY_EXISTS,
                INDEX_ALREADY_EXISTS,
            ):
                stdout.write(f"{str(e)}\n")
            else:
                raise

    async def _create_node_fulltext_index(
        self,
        target_cls: Any,
        property_name: str,
        stdout: TextIO,
        fulltext_index: FulltextIndex,
        quiet: bool,
    ) -> None:
        if await self.version_is_higher_than("5.16"):
            label = target_cls.__label__
            index_name = f"fulltext_index_{label}_{property_name}"
            if not quiet:
                stdout.write(
                    f" + Creating fulltext index for {property_name} on label {target_cls.__label__} for class {target_cls.__module__}.{target_cls.__name__}\n"
                )
            query = f"""
                CREATE FULLTEXT INDEX {index_name} FOR (n:{label}) ON EACH [n.{property_name}]
                OPTIONS {{
                    indexConfig: {{
                        `fulltext.analyzer`: '{fulltext_index.analyzer}',
                        `fulltext.eventually_consistent`: {fulltext_index.eventually_consistent}
                    }}
                }};
            """
            try:
                await self.cypher_query(query)
            except ClientError as e:
                if e.code in (
                    RULE_ALREADY_EXISTS,
                    INDEX_ALREADY_EXISTS,
                ):
                    stdout.write(f"{str(e)}\n")
                else:
                    raise
        else:
            raise FeatureNotSupported(
                f"Creation of full-text indexes from neomodel is not supported for Neo4j in version {await self.database_version}. Please upgrade to Neo4j 5.16 or higher."
            )

    async def _create_node_vector_index(
        self,
        target_cls: Any,
        property_name: str,
        stdout: TextIO,
        vector_index: VectorIndex,
        quiet: bool,
    ) -> None:
        if await self.version_is_higher_than("5.15"):
            label = target_cls.__label__
            index_name = f"vector_index_{label}_{property_name}"
            if not quiet:
                stdout.write(
                    f" + Creating vector index for {property_name} on label {label} for class {target_cls.__module__}.{target_cls.__name__}\n"
                )
            query = f"""
                CREATE VECTOR INDEX {index_name} FOR (n:{label}) ON n.{property_name}
                OPTIONS {{
                    indexConfig: {{
                        `vector.dimensions`: {vector_index.dimensions},
                        `vector.similarity_function`: '{vector_index.similarity_function}'
                    }}
                }};
            """
            try:
                await self.cypher_query(query)
            except ClientError as e:
                if e.code in (
                    RULE_ALREADY_EXISTS,
                    INDEX_ALREADY_EXISTS,
                ):
                    stdout.write(f"{str(e)}\n")
                else:
                    raise
        else:
            raise FeatureNotSupported(
                f"Creation of vector indexes from neomodel is not supported for Neo4j in version {await self.database_version}. Please upgrade to Neo4j 5.15 or higher."
            )

    async def _create_node_constraint(
        self, target_cls: Any, property_name: str, stdout: TextIO, quiet: bool
    ) -> None:
        label = target_cls.__label__
        constraint_name = f"constraint_unique_{label}_{property_name}"
        if not quiet:
            stdout.write(
                f" + Creating node unique constraint for {property_name} on label {target_cls.__label__} for class {target_cls.__module__}.{target_cls.__name__}\n"
            )
        try:
            await self.cypher_query(
                f"""CREATE CONSTRAINT {constraint_name}
                            FOR (n:{label}) REQUIRE n.{property_name} IS UNIQUE"""
            )
        except ClientError as e:
            if e.code in (
                RULE_ALREADY_EXISTS,
                CONSTRAINT_ALREADY_EXISTS,
            ):
                stdout.write(f"{str(e)}\n")
            else:
                raise

    async def _create_relationship_index(
        self,
        relationship_type: str,
        target_cls: Any,
        relationship_cls: Any,
        property_name: str,
        stdout: TextIO,
        quiet: bool,
    ) -> None:
        index_name = f"index_{relationship_type}_{property_name}"
        if not quiet:
            stdout.write(
                f" + Creating relationship index for {property_name} on relationship type {relationship_type} for relationship model {target_cls.__module__}.{relationship_cls.__name__}\n"
            )
        try:
            await self.cypher_query(
                f"CREATE INDEX {index_name} FOR ()-[r:{relationship_type}]-() ON (r.{property_name}); "
            )
        except ClientError as e:
            if e.code in (
                RULE_ALREADY_EXISTS,
                INDEX_ALREADY_EXISTS,
            ):
                stdout.write(f"{str(e)}\n")
            else:
                raise

    async def _create_relationship_fulltext_index(
        self,
        relationship_type: str,
        target_cls: Any,
        relationship_cls: Any,
        property_name: str,
        stdout: TextIO,
        fulltext_index: FulltextIndex,
        quiet: bool,
    ) -> None:
        if await self.version_is_higher_than("5.16"):
            index_name = f"fulltext_index_{relationship_type}_{property_name}"
            if not quiet:
                stdout.write(
                    f" + Creating fulltext index for {property_name} on relationship type {relationship_type} for relationship model {target_cls.__module__}.{relationship_cls.__name__}\n"
                )
            query = f"""
                CREATE FULLTEXT INDEX {index_name} FOR ()-[r:{relationship_type}]-() ON EACH [r.{property_name}]
                OPTIONS {{
                    indexConfig: {{
                        `fulltext.analyzer`: '{fulltext_index.analyzer}',
                        `fulltext.eventually_consistent`: {fulltext_index.eventually_consistent}
                    }}
                }};
            """
            try:
                await self.cypher_query(query)
            except ClientError as e:
                if e.code in (
                    RULE_ALREADY_EXISTS,
                    INDEX_ALREADY_EXISTS,
                ):
                    stdout.write(f"{str(e)}\n")
                else:
                    raise
        else:
            raise FeatureNotSupported(
                f"Creation of full-text indexes from neomodel is not supported for Neo4j in version {await self.database_version}. Please upgrade to Neo4j 5.16 or higher."
            )

    async def _create_relationship_vector_index(
        self,
        relationship_type: str,
        target_cls: Any,
        relationship_cls: Any,
        property_name: str,
        stdout: TextIO,
        vector_index: VectorIndex,
        quiet: bool,
    ) -> None:
        if await self.version_is_higher_than("5.18"):
            index_name = f"vector_index_{relationship_type}_{property_name}"
            if not quiet:
                stdout.write(
                    f" + Creating vector index for {property_name} on relationship type {relationship_type} for relationship model {target_cls.__module__}.{relationship_cls.__name__}\n"
                )
            query = f"""
                CREATE VECTOR INDEX {index_name} FOR ()-[r:{relationship_type}]-() ON r.{property_name}
                OPTIONS {{
                    indexConfig: {{
                        `vector.dimensions`: {vector_index.dimensions},
                        `vector.similarity_function`: '{vector_index.similarity_function}'
                    }}
                }};
            """
            try:
                await self.cypher_query(query)
            except ClientError as e:
                if e.code in (
                    RULE_ALREADY_EXISTS,
                    INDEX_ALREADY_EXISTS,
                ):
                    stdout.write(f"{str(e)}\n")
                else:
                    raise
        else:
            raise FeatureNotSupported(
                f"Creation of vector indexes for relationships from neomodel is not supported for Neo4j in version {await self.database_version}. Please upgrade to Neo4j 5.18 or higher."
            )

    async def _create_relationship_constraint(
        self,
        relationship_type: str,
        target_cls: Any,
        relationship_cls: Any,
        property_name: str,
        stdout: TextIO,
        quiet: bool,
    ) -> None:
        if await self.version_is_higher_than("5.7"):
            constraint_name = f"constraint_unique_{relationship_type}_{property_name}"
            if not quiet:
                stdout.write(
                    f" + Creating relationship unique constraint for {property_name} on relationship type {relationship_type} for relationship model {target_cls.__module__}.{relationship_cls.__name__}\n"
                )
            try:
                await self.cypher_query(
                    f"""CREATE CONSTRAINT {constraint_name} 
                                FOR ()-[r:{relationship_type}]-() REQUIRE r.{property_name} IS UNIQUE"""
                )
            except ClientError as e:
                if e.code in (
                    RULE_ALREADY_EXISTS,
                    CONSTRAINT_ALREADY_EXISTS,
                ):
                    stdout.write(f"{str(e)}\n")
                else:
                    raise
        else:
            raise FeatureNotSupported(
                f"Unique indexes on relationships are not supported in Neo4j version {await self.database_version}. Please upgrade to Neo4j 5.7 or higher."
            )

    async def _install_node(
        self, cls: Any, name: str, property: Property, quiet: bool, stdout: TextIO
    ) -> None:
        # Create indexes and constraints for node property
        db_property = property.get_db_property_name(name)
        if property.index:
            await self._create_node_index(
                target_cls=cls, property_name=db_property, stdout=stdout, quiet=quiet
            )
        elif property.unique_index:
            await self._create_node_constraint(
                target_cls=cls, property_name=db_property, stdout=stdout, quiet=quiet
            )

        if property.fulltext_index:
            await self._create_node_fulltext_index(
                target_cls=cls,
                property_name=db_property,
                stdout=stdout,
                fulltext_index=property.fulltext_index,
                quiet=quiet,
            )

        if property.vector_index:
            await self._create_node_vector_index(
                target_cls=cls,
                property_name=db_property,
                stdout=stdout,
                vector_index=property.vector_index,
                quiet=quiet,
            )

    async def _install_relationship(
        self, cls: Any, relationship: Any, quiet: bool, stdout: TextIO
    ) -> None:
        # Create indexes and constraints for relationship property
        relationship_cls = relationship.definition["model"]
        if relationship_cls is not None:
            relationship_type = relationship.definition["relation_type"]
            for prop_name, property in relationship_cls.defined_properties(
                aliases=False, rels=False
            ).items():
                db_property = property.get_db_property_name(prop_name)
                if property.index:
                    await self._create_relationship_index(
                        relationship_type=relationship_type,
                        target_cls=cls,
                        relationship_cls=relationship_cls,
                        property_name=db_property,
                        stdout=stdout,
                        quiet=quiet,
                    )
                elif property.unique_index:
                    await self._create_relationship_constraint(
                        relationship_type=relationship_type,
                        target_cls=cls,
                        relationship_cls=relationship_cls,
                        property_name=db_property,
                        stdout=stdout,
                        quiet=quiet,
                    )

                if property.fulltext_index:
                    await self._create_relationship_fulltext_index(
                        relationship_type=relationship_type,
                        target_cls=cls,
                        relationship_cls=relationship_cls,
                        property_name=db_property,
                        stdout=stdout,
                        fulltext_index=property.fulltext_index,
                        quiet=quiet,
                    )

                if property.vector_index:
                    await self._create_relationship_vector_index(
                        relationship_type=relationship_type,
                        target_cls=cls,
                        relationship_cls=relationship_cls,
                        property_name=db_property,
                        stdout=stdout,
                        vector_index=property.vector_index,
                        quiet=quiet,
                    )


# Create a singleton instance of the database object
adb = AsyncDatabase()


# Deprecated methods
async def change_neo4j_password(
    db: AsyncDatabase, user: str, new_password: str
) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.change_neo4j_password(user, new_password) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await db.change_neo4j_password(user, new_password)


async def clear_neo4j_database(
    db: AsyncDatabase, clear_constraints: bool = False, clear_indexes: bool = False
) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.clear_neo4j_database(clear_constraints, clear_indexes) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await db.clear_neo4j_database(clear_constraints, clear_indexes)


async def drop_constraints(quiet: bool = True, stdout: Optional[TextIO] = None) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.drop_constraints(quiet, stdout) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await adb.drop_constraints(quiet, stdout)


async def drop_indexes(quiet: bool = True, stdout: Optional[TextIO] = None) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.drop_indexes(quiet, stdout) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await adb.drop_indexes(quiet, stdout)


async def remove_all_labels(stdout: Optional[TextIO] = None) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.remove_all_labels(stdout) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await adb.remove_all_labels(stdout)


async def install_labels(
    cls: Any, quiet: bool = True, stdout: Optional[TextIO] = None
) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.install_labels(cls, quiet, stdout) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await adb.install_labels(cls, quiet, stdout)


async def install_all_labels(stdout: Optional[TextIO] = None) -> None:
    deprecated(
        """
        This method has been moved to the Database singleton (db for sync, adb for async).
        Please use adb.install_all_labels(stdout) instead.
        This direct call will be removed in an upcoming version.
        """
    )
    await adb.install_all_labels(stdout)


class AsyncTransactionProxy:
    bookmarks: Optional[Bookmarks] = None

    def __init__(
        self,
        db: AsyncDatabase,
        access_mode: Optional[str] = None,
        parallel_runtime: Optional[bool] = False,
    ):
        self.db: AsyncDatabase = db
        self.access_mode: Optional[str] = access_mode
        self.parallel_runtime: Optional[bool] = parallel_runtime

    @ensure_connection
    async def __aenter__(self) -> "AsyncTransactionProxy":
        if self.parallel_runtime and not await self.db.parallel_runtime_available():
            warnings.warn(
                "Parallel runtime is only available in Neo4j Enterprise Edition 5.13 and above. "
                "Reverting to default runtime.",
                UserWarning,
            )
            self.parallel_runtime = False
        self.db._parallel_runtime = self.parallel_runtime
        await self.db.begin(access_mode=self.access_mode, bookmarks=self.bookmarks)
        self.bookmarks = None
        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.db._parallel_runtime = False
        if exc_value:
            await self.db.rollback()

        if (
            exc_type is ClientError
            and exc_value.code == "Neo.ClientError.Schema.ConstraintValidationFailed"
        ):
            raise UniqueProperty(exc_value.message)

        if not exc_value:
            self.last_bookmark = await self.db.commit()

    def __call__(self, func: Callable) -> Callable:
        if AsyncUtil.is_async_code and not iscoroutinefunction(func):
            raise TypeError(NOT_COROUTINE_ERROR)

        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Callable:
            async with self:
                return await func(*args, **kwargs)

        return wrapper

    @property
    def with_bookmark(self) -> "BookmarkingAsyncTransactionProxy":
        return BookmarkingAsyncTransactionProxy(self.db, self.access_mode)


class BookmarkingAsyncTransactionProxy(AsyncTransactionProxy):
    def __call__(self, func: Callable) -> Callable:
        if AsyncUtil.is_async_code and not iscoroutinefunction(func):
            raise TypeError(NOT_COROUTINE_ERROR)

        async def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, None]:
            self.bookmarks = kwargs.pop("bookmarks", None)

            async with self:
                result = await func(*args, **kwargs)
                self.last_bookmark = None

            return result, self.last_bookmark

        return wrapper


class ImpersonationHandler:
    def __init__(self, db: AsyncDatabase, impersonated_user: str):
        self.db = db
        self.impersonated_user = impersonated_user

    def __enter__(self) -> "ImpersonationHandler":
        self.db.impersonated_user = self.impersonated_user
        return self

    def __exit__(
        self, exception_type: Any, exception_value: Any, exception_traceback: Any
    ) -> None:
        self.db.impersonated_user = None

        print("\nException type:", exception_type)
        print("\nException value:", exception_value)
        print("\nTraceback:", exception_traceback)

    def __call__(self, func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            with self:
                return func(*args, **kwargs)

        return wrapper


class NodeMeta(type):
    DoesNotExist: type[DoesNotExist]
    __required_properties__: tuple[str, ...]
    __all_properties__: tuple[tuple[str, Any], ...]
    __all_aliases__: tuple[tuple[str, Any], ...]
    __all_relationships__: tuple[tuple[str, Any], ...]
    __label__: str
    __optional_labels__: list[str]

    defined_properties: Callable[..., dict[str, Any]]

    def __new__(
        mcs: type, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> Any:
        namespace["DoesNotExist"] = type(name + "DoesNotExist", (DoesNotExist,), {})
        cls: NodeMeta = type.__new__(mcs, name, bases, namespace)
        cls.DoesNotExist._model_class = cls

        if hasattr(cls, "__abstract_node__"):
            delattr(cls, "__abstract_node__")
        else:
            if "deleted" in namespace:
                raise ValueError(
                    "Property name 'deleted' is not allowed as it conflicts with neomodel internals."
                )
            elif "id" in namespace:
                raise ValueError(
                    """
                        Property name 'id' is not allowed as it conflicts with neomodel internals.
                        Consider using 'uid' or 'identifier' as id is also a Neo4j internal.
                    """
                )
            elif "element_id" in namespace:
                raise ValueError(
                    """
                        Property name 'element_id' is not allowed as it conflicts with neomodel internals.
                        Consider using 'uid' or 'identifier' as element_id is also a Neo4j internal.
                    """
                )
            for key, value in (
                (x, y) for x, y in namespace.items() if isinstance(y, Property)
            ):
                value.name, value.owner = key, cls
                if hasattr(value, "setup") and callable(value.setup):
                    value.setup()

            # cache various groups of properies
            cls.__required_properties__ = tuple(
                name
                for name, property in cls.defined_properties(
                    aliases=False, rels=False
                ).items()
                if property.required or property.unique_index
            )
            cls.__all_properties__ = tuple(
                cls.defined_properties(aliases=False, rels=False).items()
            )
            cls.__all_aliases__ = tuple(
                cls.defined_properties(properties=False, rels=False).items()
            )
            cls.__all_relationships__ = tuple(
                cls.defined_properties(aliases=False, properties=False).items()
            )

            cls.__label__ = namespace.get("__label__", name)
            cls.__optional_labels__ = namespace.get("__optional_labels__", [])

            build_class_registry(cls)

        return cls


def build_class_registry(cls: Any) -> None:
    base_label_set = frozenset(cls.inherited_labels())
    optional_label_set = set(cls.inherited_optional_labels())

    # Construct all possible combinations of labels + optional labels
    possible_label_combinations = [
        frozenset(set(x).union(base_label_set))
        for i in range(1, len(optional_label_set) + 1)
        for x in combinations(optional_label_set, i)
    ]
    possible_label_combinations.append(base_label_set)

    for label_set in possible_label_combinations:
        if not hasattr(cls, "__target_databases__"):
            if label_set not in adb._NODE_CLASS_REGISTRY:
                adb._NODE_CLASS_REGISTRY[label_set] = cls
            else:
                raise NodeClassAlreadyDefined(
                    cls, adb._NODE_CLASS_REGISTRY, adb._DB_SPECIFIC_CLASS_REGISTRY
                )
        else:
            for database in cls.__target_databases__:
                if database not in adb._DB_SPECIFIC_CLASS_REGISTRY:
                    adb._DB_SPECIFIC_CLASS_REGISTRY[database] = {}
                if label_set not in adb._DB_SPECIFIC_CLASS_REGISTRY[database]:
                    adb._DB_SPECIFIC_CLASS_REGISTRY[database][label_set] = cls
                else:
                    raise NodeClassAlreadyDefined(
                        cls, adb._NODE_CLASS_REGISTRY, adb._DB_SPECIFIC_CLASS_REGISTRY
                    )


NodeBase: type = NodeMeta(
    "NodeBase", (AsyncPropertyManager,), {"__abstract_node__": True}
)


class AsyncStructuredNode(NodeBase):
    """
    Base class for all node definitions to inherit from.

    If you want to create your own abstract classes set:
        __abstract_node__ = True
    """

    # static properties

    __abstract_node__ = True

    # magic methods

    def __init__(self, *args: Any, **kwargs: Any):
        if "deleted" in kwargs:
            raise ValueError("deleted property is reserved for neomodel")

        for key, val in self.__all_relationships__:
            self.__dict__[key] = val.build_manager(self, key)

        super().__init__(*args, **kwargs)

    def __eq__(self, other: Any) -> bool:
        """
        Compare two node objects.
        If both nodes were saved to the database, compare them by their element_id.
        Otherwise, compare them using object id in memory.
        If `other` is not a node, always return False.
        """
        if not isinstance(other, (AsyncStructuredNode,)):
            return False
        if self.was_saved and other.was_saved:
            return self.element_id == other.element_id
        return id(self) == id(other)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return repr(self.__properties__)

    # dynamic properties

    @classproperty
    def nodes(self) -> Any:
        """
        Returns a NodeSet object representing all nodes of the classes label
        :return: NodeSet
        :rtype: NodeSet
        """
        from neomodel.async_.match import AsyncNodeSet

        return AsyncNodeSet(self)

    @property
    def element_id(self) -> Optional[Any]:
        if hasattr(self, "element_id_property"):
            return self.element_id_property
        return None

    # Version 4.4 support - id is deprecated in version 5.x
    @property
    def id(self) -> int:
        try:
            return int(self.element_id_property)
        except (TypeError, ValueError):
            raise ValueError(
                "id is deprecated in Neo4j version 5, please migrate to element_id. If you use the id in a Cypher query, replace id() by elementId()."
            )

    @property
    def was_saved(self) -> bool:
        """
        Shows status of node in the database. False, if node hasn't been saved yet, True otherwise.
        """
        return self.element_id is not None

    # methods

    @classmethod
    async def _build_merge_query(
        cls,
        merge_params: tuple[dict[str, Any], ...],
        update_existing: bool = False,
        lazy: bool = False,
        relationship: Optional[Any] = None,
    ) -> tuple[str, dict[str, Any]]:
        """
        Get a tuple of a CYPHER query and a params dict for the specified MERGE query.

        :param merge_params: The target node match parameters, each node must have a "create" key and optional "update".
        :type merge_params: list of dict
        :param update_existing: True to update properties of existing nodes, default False to keep existing values.
        :type update_existing: bool
        :rtype: tuple
        """
        query_params: dict[str, Any] = {"merge_params": merge_params}
        n_merge_labels = ":".join(cls.inherited_labels())
        n_merge_prm = ", ".join(
            (
                f"{getattr(cls, p).get_db_property_name(p)}: params.create.{getattr(cls, p).get_db_property_name(p)}"
                for p in cls.__required_properties__
            )
        )
        n_merge = f"n:{n_merge_labels} {{{n_merge_prm}}}"
        if relationship is None:
            # create "simple" unwind query
            query = f"UNWIND $merge_params as params\n MERGE ({n_merge})\n "
        else:
            # validate relationship
            if not isinstance(relationship.source, AsyncStructuredNode):
                raise ValueError(
                    f"relationship source [{repr(relationship.source)}] is not a StructuredNode"
                )
            relation_type = relationship.definition.get("relation_type")
            if not relation_type:
                raise ValueError(
                    "No relation_type is specified on provided relationship"
                )

            from neomodel.async_.match import _rel_helper

            if relationship.source.element_id is None:
                raise RuntimeError(
                    "Could not identify the relationship source, its element id was None."
                )
            query_params["source_id"] = await adb.parse_element_id(
                relationship.source.element_id
            )
            query = f"MATCH (source:{relationship.source.__label__}) WHERE {await adb.get_id_method()}(source) = $source_id\n "
            query += "WITH source\n UNWIND $merge_params as params \n "
            query += "MERGE "
            query += _rel_helper(
                lhs="source",
                rhs=n_merge,
                ident=None,
                relation_type=relation_type,
                direction=relationship.definition["direction"],
            )

        query += "ON CREATE SET n = params.create\n "
        # if update_existing, write properties on match as well
        if update_existing is True:
            query += "ON MATCH SET n += params.update\n"

        # close query
        if lazy:
            query += f"RETURN {await adb.get_id_method()}(n)"
        else:
            query += "RETURN n"

        return query, query_params

    @classmethod
    async def create(cls, *props: tuple, **kwargs: dict[str, Any]) -> list:
        """
        Call to CREATE with parameters map. A new instance will be created and saved.

        :param props: dict of properties to create the nodes.
        :type props: tuple
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :type: bool
        :rtype: list
        """

        if "streaming" in kwargs:
            warnings.warn(
                STREAMING_WARNING,
                category=DeprecationWarning,
                stacklevel=1,
            )

        lazy = kwargs.get("lazy", False)
        # create mapped query
        query = f"CREATE (n:{':'.join(cls.inherited_labels())} $create_params)"

        # close query
        if lazy:
            query += f" RETURN {await adb.get_id_method()}(n)"
        else:
            query += " RETURN n"

        results = []
        for item in [
            cls.deflate(p, obj=_UnsavedNode(), skip_empty=True) for p in props
        ]:
            node, _ = await adb.cypher_query(query, {"create_params": item})
            results.extend(node[0])

        nodes = [cls.inflate(node) for node in results]

        if not lazy and hasattr(cls, "post_create"):
            for node in nodes:
                node.post_create()

        return nodes

    @classmethod
    async def create_or_update(cls, *props: tuple, **kwargs: dict[str, Any]) -> list:
        """
        Call to MERGE with parameters map. A new instance will be created and saved if does not already exists,
        this is an atomic operation. If an instance already exists all optional properties specified will be updated.

        Note that the post_create hook isn't called after create_or_update

        :param props: List of dict arguments to get or create the entities with.
        :type props: tuple
        :param relationship: Optional, relationship to get/create on when new entity is created.
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :rtype: list
        """
        lazy: bool = bool(kwargs.get("lazy", False))
        relationship = kwargs.get("relationship")

        # build merge query, make sure to update only explicitly specified properties
        create_or_update_params = []
        for specified, deflated in [
            (p, cls.deflate(p, skip_empty=True)) for p in props
        ]:
            create_or_update_params.append(
                {
                    "create": deflated,
                    "update": dict(
                        (k, v) for k, v in deflated.items() if k in specified
                    ),
                }
            )
        query, params = await cls._build_merge_query(
            tuple(create_or_update_params),
            update_existing=True,
            relationship=relationship,
            lazy=lazy,
        )

        if "streaming" in kwargs:
            warnings.warn(
                STREAMING_WARNING,
                category=DeprecationWarning,
                stacklevel=1,
            )

        # fetch and build instance for each result
        results = await adb.cypher_query(query, params)
        return [cls.inflate(r[0]) for r in results[0]]

    async def cypher(
        self, query: str, params: Optional[dict[str, Any]] = None
    ) -> tuple[Optional[list], Optional[tuple[str, ...]]]:
        """
        Execute a cypher query with the param 'self' pre-populated with the nodes neo4j id.

        :param query: cypher query string
        :type: string
        :param params: query parameters
        :type: dict
        :return: tuple containing a list of query results, and the meta information as a tuple
        :rtype: tuple
        """
        self._pre_action_check("cypher")
        _params = params or {}
        if self.element_id is None:
            raise ValueError("Can't run cypher operation on unsaved node")
        element_id = await adb.parse_element_id(self.element_id)
        _params.update({"self": element_id})
        return await adb.cypher_query(query, _params)

    @hooks
    async def delete(self) -> bool:
        """
        Delete a node and its relationships

        :return: True
        """
        self._pre_action_check("delete")
        await self.cypher(
            f"MATCH (self) WHERE {await adb.get_id_method()}(self)=$self DETACH DELETE self"
        )
        delattr(self, "element_id_property")
        self.deleted = True
        return True

    @classmethod
    async def get_or_create(cls: Any, *props: tuple, **kwargs: dict[str, Any]) -> list:
        """
        Call to MERGE with parameters map. A new instance will be created and saved if does not already exist,
        this is an atomic operation.
        Parameters must contain all required properties, any non required properties with defaults will be generated.

        Note that the post_create hook isn't called after get_or_create

        :param props: Arguments to get_or_create as tuple of dict with property names and values to get or create
                      the entities with.
        :type props: tuple
        :param relationship: Optional, relationship to get/create on when new entity is created.
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :rtype: list
        """
        lazy = kwargs.get("lazy", False)
        relationship = kwargs.get("relationship")

        # build merge query
        get_or_create_params = [
            {"create": cls.deflate(p, skip_empty=True)} for p in props
        ]
        query, params = await cls._build_merge_query(
            tuple(get_or_create_params), relationship=relationship, lazy=lazy
        )

        if "streaming" in kwargs:
            warnings.warn(
                STREAMING_WARNING,
                category=DeprecationWarning,
                stacklevel=1,
            )

        # fetch and build instance for each result
        results = await adb.cypher_query(query, params)
        return [cls.inflate(r[0]) for r in results[0]]

    @classmethod
    def inflate(cls: Any, node: Any) -> Any:
        """
        Inflate a raw neo4j_driver node to a neomodel node
        :param node:
        :return: node object
        """
        # support lazy loading
        if isinstance(node, str) or isinstance(node, int):
            snode = cls()
            snode.element_id_property = node
        else:
            snode = super().inflate(node)
            snode.element_id_property = node.element_id

        return snode

    @classmethod
    def inherited_labels(cls: Any) -> list[str]:
        """
        Return list of labels from nodes class hierarchy.

        :return: list
        """
        return [
            scls.__label__
            for scls in cls.mro()
            if hasattr(scls, "__label__") and not hasattr(scls, "__abstract_node__")
        ]

    @classmethod
    def inherited_optional_labels(cls: Any) -> list[str]:
        """
        Return list of optional labels from nodes class hierarchy.

        :return: list
        :rtype: list
        """
        return [
            label
            for scls in cls.mro()
            for label in getattr(scls, "__optional_labels__", [])
            if not hasattr(scls, "__abstract_node__")
        ]

    async def labels(self) -> list[str]:
        """
        Returns list of labels tied to the node from neo4j.

        :return: list of labels
        :rtype: list
        """
        self._pre_action_check("labels")
        result = await self.cypher(
            f"MATCH (n) WHERE {await adb.get_id_method()}(n)=$self " "RETURN labels(n)"
        )
        if result is None or result[0] is None:
            raise ValueError("Could not get labels, node may not exist")
        return result[0][0][0]

    def _pre_action_check(self, action: str) -> None:
        if hasattr(self, "deleted") and self.deleted:
            raise ValueError(
                f"{self.__class__.__name__}.{action}() attempted on deleted node"
            )
        if not hasattr(self, "element_id"):
            raise ValueError(
                f"{self.__class__.__name__}.{action}() attempted on unsaved node"
            )

    async def refresh(self) -> None:
        """
        Reload the node from neo4j
        """
        self._pre_action_check("refresh")
        if hasattr(self, "element_id"):
            results = await self.cypher(
                f"MATCH (n) WHERE {await adb.get_id_method()}(n)=$self RETURN n"
            )
            request = results[0]
            if not request or not request[0]:
                raise self.__class__.DoesNotExist("Can't refresh non existent node")
            node = self.inflate(request[0][0])
            for key, val in node.__properties__.items():
                setattr(self, key, val)
        else:
            raise ValueError("Can't refresh unsaved node")

    @hooks
    async def save(self) -> "AsyncStructuredNode":
        """
        Save the node to neo4j or raise an exception

        :return: the node instance
        """

        # create or update instance node
        if hasattr(self, "element_id_property"):
            # update
            params = self.deflate(self.__properties__, self)
            query = f"MATCH (n) WHERE {await adb.get_id_method()}(n)=$self\n"

            if params:
                query += "SET "
                query += ",\n".join([f"n.{key} = ${key}" for key in params])
                query += "\n"
            if self.inherited_labels():
                query += "\n".join(
                    [f"SET n:`{label}`" for label in self.inherited_labels()]
                )
            await self.cypher(query, params)
        elif hasattr(self, "deleted") and self.deleted:
            raise ValueError(
                f"{self.__class__.__name__}.save() attempted on deleted node"
            )
        else:  # create
            result = await self.create(self.__properties__)
            created_node = result[0]
            self.element_id_property = created_node.element_id
        return self



from typing import Any, Optional

from neo4j.graph import Relationship

from neomodel.hooks import hooks
from neomodel.properties import Property
from neomodel.sync_.core import db
from neomodel.sync_.property_manager import PropertyManager

ELEMENT_ID_MIGRATION_NOTICE = "id is deprecated in Neo4j version 5, please migrate to element_id. If you use the id in a Cypher query, replace id() by elementId()."


class RelationshipMeta(type):
    def __new__(
        mcs: type, name: str, bases: tuple[type, ...], dct: dict[str, Any]
    ) -> Any:
        inst: RelationshipMeta = type.__new__(mcs, name, bases, dct)
        for key, value in dct.items():
            if issubclass(value.__class__, Property):
                if key == "source" or key == "target":
                    raise ValueError(
                        "Property names 'source' and 'target' are not allowed as they conflict with neomodel internals."
                    )
                elif key == "id":
                    raise ValueError(
                        """
                            Property name 'id' is not allowed as it conflicts with neomodel internals.
                            Consider using 'uid' or 'identifier' as id is also a Neo4j internal.
                        """
                    )
                elif key == "element_id":
                    raise ValueError(
                        """
                            Property name 'element_id' is not allowed as it conflicts with neomodel internals.
                            Consider using 'uid' or 'identifier' as element_id is also a Neo4j internal.
                        """
                    )
                value.name = key
                value.owner = inst

                # support for 'magic' properties
                if hasattr(value, "setup") and hasattr(value.setup, "__call__"):
                    value.setup()
        return inst


class StructuredRelBase(PropertyManager, metaclass=RelationshipMeta):
    pass


class StructuredRel(StructuredRelBase):
    """
    Base class for relationship objects
    """

    def __init__(self, *args: Any, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)

    @property
    def element_id(self) -> Optional[Any]:
        if hasattr(self, "element_id_property"):
            return self.element_id_property
        return None

    @property
    def _start_node_element_id(self) -> Optional[Any]:
        if hasattr(self, "_start_node_element_id_property"):
            return self._start_node_element_id_property
        return None

    @property
    def _end_node_element_id(self) -> Optional[Any]:
        if hasattr(self, "_end_node_element_id_property"):
            return self._end_node_element_id_property
        return None

    # Version 4.4 support - id is deprecated in version 5.x
    @property
    def id(self) -> int:
        try:
            return int(self.element_id_property)
        except (TypeError, ValueError) as exc:
            raise ValueError(ELEMENT_ID_MIGRATION_NOTICE) from exc

    # Version 4.4 support - id is deprecated in version 5.x
    @property
    def _start_node_id(self) -> int:
        try:
            return int(self._start_node_element_id_property)
        except (TypeError, ValueError) as exc:
            raise ValueError(ELEMENT_ID_MIGRATION_NOTICE) from exc

    # Version 4.4 support - id is deprecated in version 5.x
    @property
    def _end_node_id(self) -> int:
        try:
            return int(self._end_node_element_id_property)
        except (TypeError, ValueError) as exc:
            raise ValueError(ELEMENT_ID_MIGRATION_NOTICE) from exc

    @hooks
    def save(self) -> "StructuredRel":
        """
        Save the relationship

        :return: self
        """
        props = self.deflate(self.__properties__)
        query = f"MATCH ()-[r]->() WHERE {db.get_id_method()}(r)=$self "
        query += "".join([f" SET r.{key} = ${key}" for key in props])
        props["self"] = db.parse_element_id(self.element_id)

        db.cypher_query(query, props)

        return self

    def start_node(self) -> Any:
        """
        Get start node

        :return: StructuredNode
        """
        results = db.cypher_query(
            f"""
            MATCH (aNode)
            WHERE {db.get_id_method()}(aNode)=$start_node_element_id
            RETURN aNode
            """,
            {"start_node_element_id": db.parse_element_id(self._start_node_element_id)},
            resolve_objects=True,
        )
        if results is None or results[0] is None or results[0][0] is None:
            raise ValueError(
                f"Start node with elementId {self._start_node_element_id} not found"
            )
        return results[0][0][0]

    def end_node(self) -> Any:
        """
        Get end node

        :return: StructuredNode
        """
        results = db.cypher_query(
            f"""
            MATCH (aNode)
            WHERE {db.get_id_method()}(aNode)=$end_node_element_id
            RETURN aNode
            """,
            {"end_node_element_id": db.parse_element_id(self._end_node_element_id)},
            resolve_objects=True,
        )
        if results is None or results[0] is None or results[0][0] is None:
            raise ValueError(
                f"Start node with elementId {self._start_node_element_id} not found"
            )
        return results[0][0][0]

    @classmethod
    def inflate(cls: Any, rel: Relationship) -> "StructuredRel":
        """
        Inflate a neo4j_driver relationship object to a neomodel object
        :param rel:
        :return: StructuredRel
        """
        srel = super().inflate(rel)
        if rel.start_node is not None:
            srel._start_node_element_id_property = rel.start_node.element_id
        if rel.end_node is not None:
            srel._end_node_element_id_property = rel.end_node.element_id
        srel.element_id_property = rel.element_id
        return srel



import builtins
from test._async_compat import mark_async_test

import pytest
from neo4j.exceptions import ClientError as CypherError
from numpy import ndarray
from pandas import DataFrame, Series

from neomodel import AsyncStructuredNode, StringProperty, adb
from neomodel._async_compat.util import AsyncUtil


class User2(AsyncStructuredNode):
    name = StringProperty()
    email = StringProperty()


class UserPandas(AsyncStructuredNode):
    name = StringProperty()
    email = StringProperty()


class UserNP(AsyncStructuredNode):
    name = StringProperty()
    email = StringProperty()


@pytest.fixture
def hide_available_pkg(monkeypatch, request):
    import_orig = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name == request.param:
            raise ImportError()
        return import_orig(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", mocked_import)


@mark_async_test
async def test_cypher():
    """
    test result format is backward compatible with earlier versions of neomodel
    """

    jim = await User2(email="jim1@test.com").save()
    data, meta = await jim.cypher(
        f"MATCH (a) WHERE {await adb.get_id_method()}(a)=$self RETURN a.email"
    )
    assert data[0][0] == "jim1@test.com"
    assert "a.email" in meta

    data, meta = await jim.cypher(
        f"""
            MATCH (a) WHERE {await adb.get_id_method()}(a)=$self
            MATCH (a)<-[:USER2]-(b)
            RETURN a, b, 3
        """
    )
    assert "a" in meta and "b" in meta


@mark_async_test
async def test_cypher_syntax_error():
    jim = await User2(email="jim1@test.com").save()
    try:
        await jim.cypher(
            f"MATCH a WHERE {await adb.get_id_method()}(a)={{self}} RETURN xx"
        )
    except CypherError as e:
        assert hasattr(e, "message")
        assert hasattr(e, "code")
    else:
        assert False, "CypherError not raised."


@mark_async_test
@pytest.mark.parametrize("hide_available_pkg", ["pandas"], indirect=True)
async def test_pandas_not_installed(hide_available_pkg):
    # We run only the async version, because this fails on second run
    # because import error is thrown only when pandas.py is imported
    if not AsyncUtil.is_async_code:
        pytest.skip("This test is async only")
    with pytest.raises(ImportError):
        with pytest.warns(
            UserWarning,
            match="The neomodel.integration.pandas module expects pandas to be installed",
        ):
            from neomodel.integration.pandas import to_dataframe

            _ = to_dataframe(await adb.cypher_query("MATCH (a) RETURN a.name AS name"))


@mark_async_test
async def test_pandas_integration():
    from neomodel.integration.pandas import to_dataframe, to_series

    jimla = await UserPandas(email="jimla@test.com", name="jimla").save()
    jimlo = await UserPandas(email="jimlo@test.com", name="jimlo").save()

    # Test to_dataframe
    df = to_dataframe(
        await adb.cypher_query(
            "MATCH (a:UserPandas) RETURN a.name AS name, a.email AS email ORDER BY name"
        )
    )

    assert isinstance(df, DataFrame)
    assert df.shape == (2, 2)
    assert df["name"].tolist() == ["jimla", "jimlo"]

    # Also test passing an index and dtype to to_dataframe
    df = to_dataframe(
        await adb.cypher_query(
            "MATCH (a:UserPandas) RETURN a.name AS name, a.email AS email ORDER BY name"
        ),
        index=df["email"],
        dtype=str,
    )

    assert df.index.inferred_type == "string"

    # Next test to_series
    series = to_series(
        await adb.cypher_query(
            "MATCH (a:UserPandas) RETURN a.name AS name ORDER BY name"
        )
    )

    assert isinstance(series, Series)
    assert series.shape == (2,)
    assert df["name"].tolist() == ["jimla", "jimlo"]


@mark_async_test
@pytest.mark.parametrize("hide_available_pkg", ["numpy"], indirect=True)
async def test_numpy_not_installed(hide_available_pkg):
    # We run only the async version, because this fails on second run
    # because import error is thrown only when numpy.py is imported
    if not AsyncUtil.is_async_code:
        pytest.skip("This test is async only")
    with pytest.raises(ImportError):
        with pytest.warns(
            UserWarning,
            match="The neomodel.integration.numpy module expects numpy to be installed",
        ):
            from neomodel.integration.numpy import to_ndarray

            _ = to_ndarray(
                await adb.cypher_query("MATCH (a) RETURN a.name AS name ORDER BY name")
            )


@mark_async_test
async def test_numpy_integration():
    from neomodel.integration.numpy import to_ndarray

    jimly = await UserNP(email="jimly@test.com", name="jimly").save()
    jimlu = await UserNP(email="jimlu@test.com", name="jimlu").save()

    array = to_ndarray(
        await adb.cypher_query(
            "MATCH (a:UserNP) RETURN a.name AS name, a.email AS email ORDER BY name"
        )
    )

    assert isinstance(array, ndarray)
    assert array.shape == (2, 2)
    assert array[0][0] == "jimlu"



import inspect
import re
import string
import warnings
from dataclasses import dataclass
from typing import Any, Iterator
from typing import Optional as TOptional
from typing import Tuple, Union

from neomodel.exceptions import MultipleNodesReturned
from neomodel.match_q import Q, QBase
from neomodel.properties import AliasProperty, ArrayProperty, Property
from neomodel.sync_ import relationship_manager
from neomodel.sync_.core import StructuredNode, db
from neomodel.sync_.relationship import StructuredRel
from neomodel.semantic_filters import VectorFilter
from neomodel.typing import Subquery, Transformation
from neomodel.util import INCOMING, OUTGOING

CYPHER_ACTIONS_WITH_SIDE_EFFECT_EXPR = re.compile(r"(?i:MERGE|CREATE|DELETE|DETACH)")


def _rel_helper(
    lhs: str,
    rhs: str,
    ident: TOptional[str] = None,
    relation_type: TOptional[str] = None,
    direction: TOptional[int] = None,
    relation_properties: TOptional[dict] = None,
    **kwargs: dict[str, Any],  # NOSONAR
) -> str:
    """
    Generate a relationship matching string, with specified parameters.
    Examples:
    relation_direction = OUTGOING: (lhs)-[relation_ident:relation_type]->(rhs)
    relation_direction = INCOMING: (lhs)<-[relation_ident:relation_type]-(rhs)
    relation_direction = EITHER: (lhs)-[relation_ident:relation_type]-(rhs)

    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param relation_properties: dictionary of relationship properties to match
    :returns: string
    """
    rel_props = ""

    if relation_properties:
        rel_props_str = ", ".join(
            (f"{key}: {value}" for key, value in relation_properties.items())
        )
        rel_props = f" {{{rel_props_str}}}"

    rel_def = ""
    # relation_type is unspecified
    if relation_type is None:
        rel_def = ""
    # all("*" wildcard) relation_type
    elif relation_type == "*":
        rel_def = "[*]"
    else:
        # explicit relation_type
        rel_def = f"[{ident if ident else ''}:`{relation_type}`{rel_props}]"

    stmt = ""
    if direction == OUTGOING:
        stmt = f"-{rel_def}->"
    elif direction == INCOMING:
        stmt = f"<-{rel_def}-"
    else:
        stmt = f"-{rel_def}-"

    # Make sure not to add parenthesis when they are already present
    if lhs[-1] != ")":
        lhs = f"({lhs})"
    if rhs[-1] != ")":
        rhs = f"({rhs})"

    return f"{lhs}{stmt}{rhs}"


def _rel_merge_helper(
    lhs: str,
    rhs: str,
    ident: str = "neomodelident",
    relation_type: TOptional[str] = None,
    direction: TOptional[int] = None,
    relation_properties: TOptional[dict] = None,
    **kwargs: dict[str, Any],  # NOSONAR
) -> str:
    """
    Generate a relationship merging string, with specified parameters.
    Examples:
    relation_direction = OUTGOING: (lhs)-[relation_ident:relation_type]->(rhs)
    relation_direction = INCOMING: (lhs)<-[relation_ident:relation_type]-(rhs)
    relation_direction = EITHER: (lhs)-[relation_ident:relation_type]-(rhs)

    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param relation_properties: dictionary of relationship properties to merge
    :returns: string
    """

    if direction == OUTGOING:
        stmt = "-{0}->"
    elif direction == INCOMING:
        stmt = "<-{0}-"
    else:
        stmt = "-{0}-"

    rel_props = ""
    rel_none_props = ""

    if relation_properties:
        rel_props_str = ", ".join(
            (
                f"{key}: {value}"
                for key, value in relation_properties.items()
                if value is not None
            )
        )
        rel_props = f" {{{rel_props_str}}}"
        if None in relation_properties.values():
            rel_prop_val_str = ", ".join(
                (
                    f"{ident}.{key}=${key!s}"
                    for key, value in relation_properties.items()
                    if value is None
                )
            )
            rel_none_props = (
                f" ON CREATE SET {rel_prop_val_str} ON MATCH SET {rel_prop_val_str}"
            )
    # relation_type is unspecified
    if relation_type is None:
        stmt = stmt.format("")
    # all("*" wildcard) relation_type
    elif relation_type == "*":
        stmt = stmt.format("[*]")
    else:
        # explicit relation_type
        stmt = stmt.format(f"[{ident}:`{relation_type}`{rel_props}]")

    return f"({lhs}){stmt}({rhs}){rel_none_props}"


# special operators
_SPECIAL_OPERATOR_IN = "IN"
_SPECIAL_OPERATOR_ARRAY_IN = "any(x IN {ident}.{prop} WHERE x IN {val})"
_SPECIAL_OPERATOR_INSENSITIVE = "(?i)"
_SPECIAL_OPERATOR_ISNULL = "IS NULL"
_SPECIAL_OPERATOR_ISNOTNULL = "IS NOT NULL"
_SPECIAL_OPERATOR_REGEX = "=~"

_UNARY_OPERATORS = (_SPECIAL_OPERATOR_ISNULL, _SPECIAL_OPERATOR_ISNOTNULL)

_REGEX_INSENSITIVE = _SPECIAL_OPERATOR_INSENSITIVE + "{}"
_REGEX_CONTAINS = ".*{}.*"
_REGEX_STARTSWITH = "{}.*"
_REGEX_ENDSWITH = ".*{}"

# regex operations that require escaping
_STRING_REGEX_OPERATOR_TABLE = {
    "iexact": _REGEX_INSENSITIVE,
    "contains": _REGEX_CONTAINS,
    "icontains": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_CONTAINS,
    "startswith": _REGEX_STARTSWITH,
    "istartswith": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_STARTSWITH,
    "endswith": _REGEX_ENDSWITH,
    "iendswith": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_ENDSWITH,
}
# regex operations that do not require escaping
_REGEX_OPERATOR_TABLE = {
    "iregex": _REGEX_INSENSITIVE,
}
# list all regex operations, these will require formatting of the value
_REGEX_OPERATOR_TABLE.update(_STRING_REGEX_OPERATOR_TABLE)

# list all supported operators
OPERATOR_TABLE = {
    "lt": "<",
    "gt": ">",
    "lte": "<=",
    "gte": ">=",
    "ne": "<>",
    "in": _SPECIAL_OPERATOR_IN,
    "isnull": _SPECIAL_OPERATOR_ISNULL,
    "regex": _SPECIAL_OPERATOR_REGEX,
    "exact": "=",
}
# add all regex operators
OPERATOR_TABLE.update(_REGEX_OPERATOR_TABLE)

path_split_regex = re.compile(r"__(?!_)|\|")


def install_traversals(cls: type[StructuredNode], node_set: "NodeSet") -> None:
    """
    For a StructuredNode class install Traversal objects for each
    relationship definition on a NodeSet instance
    """
    rels = cls.defined_properties(rels=True, aliases=False, properties=False)

    for key in rels.keys():
        if hasattr(node_set, key):
            raise ValueError(f"Cannot install traversal '{key}' exists on NodeSet")

        rel = getattr(cls, key)
        rel.lookup_node_class()

        traversal = Traversal(source=node_set, name=key, definition=rel.definition)
        setattr(node_set, key, traversal)


def _handle_special_operators(
    property_obj: Property, key: str, value: str, operator: str, prop: str
) -> Tuple[str, str, str]:
    if operator == _SPECIAL_OPERATOR_IN:
        if not isinstance(value, (list, tuple)):
            raise ValueError(
                f"Value must be a tuple or list for IN operation {key}={value}"
            )
        if isinstance(property_obj, ArrayProperty):
            deflated_value = property_obj.deflate(value)
            operator = _SPECIAL_OPERATOR_ARRAY_IN
        else:
            deflated_value = [property_obj.deflate(v) for v in value]
    elif operator == _SPECIAL_OPERATOR_ISNULL:
        if not isinstance(value, bool):
            raise ValueError(f"Value must be a bool for isnull operation on {key}")
        operator = "IS NULL" if value else "IS NOT NULL"
        deflated_value = None
    elif operator in _REGEX_OPERATOR_TABLE.values():
        deflated_value = property_obj.deflate(value)
        if not isinstance(deflated_value, str):
            raise ValueError(f"Must be a string value for {key}")
        if operator in _STRING_REGEX_OPERATOR_TABLE.values():
            deflated_value = re.escape(deflated_value)
        deflated_value = operator.format(deflated_value)
        operator = _SPECIAL_OPERATOR_REGEX
    else:
        deflated_value = property_obj.deflate(value)

    return deflated_value, operator, prop


def _deflate_value(
    cls: type[StructuredNode],
    property_obj: Property,
    key: str,
    value: str,
    operator: str,
    prop: str,
) -> Tuple[str, str, str]:
    if isinstance(property_obj, AliasProperty):
        prop = property_obj.aliased_to()
        deflated_value = getattr(cls, prop).deflate(value)
    else:
        # handle special operators
        deflated_value, operator, prop = _handle_special_operators(
            property_obj, key, value, operator, prop
        )

    return deflated_value, operator, prop


def _initialize_filter_args_variables(
    cls: type[StructuredNode], key: str
) -> Tuple[type[StructuredNode], None, None, str, bool, str]:
    current_class = cls
    current_rel_model = None
    leaf_prop = None
    operator = "="
    is_rel_property = "|" in key
    prop = key

    return (
        current_class,
        current_rel_model,
        leaf_prop,
        operator,
        is_rel_property,
        prop,
    )


def _process_filter_key(
    cls: type[StructuredNode], key: str
) -> Tuple[Property, str, str]:
    (
        current_class,
        current_rel_model,
        leaf_prop,
        operator,
        is_rel_property,
        prop,
    ) = _initialize_filter_args_variables(cls, key)

    for part in re.split(path_split_regex, key):
        defined_props = current_class.defined_properties(rels=True)
        # update defined props dictionary with relationship properties if
        # we are filtering by property
        if is_rel_property and current_rel_model:
            defined_props.update(current_rel_model.defined_properties(rels=True))
        if part in defined_props:
            if isinstance(
                defined_props[part], relationship_manager.RelationshipDefinition
            ):
                defined_props[part].lookup_node_class()
                current_class = defined_props[part].definition["node_class"]
                current_rel_model = defined_props[part].definition["model"]
        elif part in OPERATOR_TABLE:
            operator = OPERATOR_TABLE[part]
            prop, _ = prop.rsplit("__", 1)
            continue
        else:
            raise ValueError(
                f"No such property {part} on {cls.__name__}. Note that Neo4j internals like id or element_id are not allowed for use in this operation."
            )
        leaf_prop = part

    if leaf_prop is None:
        raise ValueError(f"Badly formed filter, no property found in {key}")
    if is_rel_property and current_rel_model:
        property_obj = getattr(current_rel_model, leaf_prop)
    else:
        property_obj = getattr(current_class, leaf_prop)

    return property_obj, operator, prop


def process_filter_args(cls: type[StructuredNode], kwargs: dict[str, Any]) -> dict:
    """
    loop through properties in filter parameters check they match class definition
    deflate them and convert into something easy to generate cypher from
    """
    output = {}

    for key, value in kwargs.items():
        property_obj, operator, prop = _process_filter_key(cls, key)
        deflated_value, operator, prop = _deflate_value(
            cls, property_obj, key, value, operator, prop
        )
        # map property to correct property name in the database
        db_property = prop

        output[db_property] = (operator, deflated_value)
    return output


def process_has_args(
    cls: type[StructuredNode], kwargs: dict[str, Any]
) -> tuple[dict, dict]:
    """
    loop through has parameters check they correspond to class rels defined
    """
    rel_definitions = cls.defined_properties(properties=False, rels=True, aliases=False)

    match, dont_match = {}, {}

    for key, value in kwargs.items():
        if key not in rel_definitions:
            raise ValueError(f"No such relation {key} defined on a {cls.__name__}")

        rhs_ident = key

        rel_definitions[key].lookup_node_class()

        if value is True:
            match[rhs_ident] = rel_definitions[key].definition
        elif value is False:
            dont_match[rhs_ident] = rel_definitions[key].definition
        elif isinstance(value, NodeSet):
            raise NotImplementedError("Not implemented yet")
        else:
            raise ValueError("Expecting True / False / NodeSet got: " + repr(value))

    return match, dont_match


class QueryAST:
    match: list[str]
    optional_match: list[str]
    where: list[str]
    with_clause: TOptional[str]
    return_clause: TOptional[str]
    order_by: TOptional[list[str]]
    skip: TOptional[int]
    limit: TOptional[int]
    result_class: TOptional[type]
    lookup: TOptional[str]
    additional_return: TOptional[list[str]]
    is_count: TOptional[bool]
    vector_index_query: TOptional[type]

    def __init__(
        self,
        match: TOptional[list[str]] = None,
        optional_match: TOptional[list[str]] = None,
        where: TOptional[list[str]] = None,
        optional_where: TOptional[list[str]] = None,
        with_clause: TOptional[str] = None,
        return_clause: TOptional[str] = None,
        order_by: TOptional[list[str]] = None,
        skip: TOptional[int] = None,
        limit: TOptional[int] = None,
        result_class: TOptional[type] = None,
        lookup: TOptional[str] = None,
        additional_return: TOptional[list[str]] = None,
        is_count: TOptional[bool] = False,
        vector_index_query: TOptional[type] = None,
    ) -> None:
        self.match = match if match else []
        self.optional_match = optional_match if optional_match else []
        self.where = where if where else []
        self.optional_where = optional_where if optional_where else []
        self.with_clause = with_clause
        self.return_clause = return_clause
        self.order_by = order_by
        self.skip = skip
        self.limit = limit
        self.result_class = result_class
        self.lookup = lookup
        self.additional_return: list[str] = (
            additional_return if additional_return else []
        )
        self.is_count = is_count
        self.vector_index_query = vector_index_query
        self.subgraph: dict = {}
        self.mixed_filters: bool = False

class QueryBuilder:
    def __init__(
        self, node_set: "BaseSet", subquery_namespace: TOptional[str] = None
    ) -> None:
        self.node_set = node_set
        self._ast = QueryAST()
        self._query_params: dict = {}
        self._place_holder_registry: dict = {}
        self._relation_identifier_count: int = 0
        self._node_identifier_count: int = 0
        self._subquery_namespace: TOptional[str] = subquery_namespace

    def build_ast(self) -> "QueryBuilder":
        if isinstance(self.node_set, NodeSet) and hasattr(
            self.node_set, "relations_to_fetch"
        ):
            for relation in self.node_set.relations_to_fetch:
                self.build_traversal_from_path(relation, self.node_set.source)
        if isinstance(self.node_set, NodeSet) and hasattr(
            self.node_set, "_vector_query"
        ): 
            if self.node_set._vector_query:
                self.build_vector_query(self.node_set._vector_query, self.node_set.source)

        self.build_source(self.node_set)

        if hasattr(self.node_set, "skip"):
            self._ast.skip = self.node_set.skip
        if hasattr(self.node_set, "limit"):
            self._ast.limit = self.node_set.limit

        return self

    def build_source(
        self, source: Union["Traversal", "NodeSet", StructuredNode, Any]
    ) -> str:
        if isinstance(source, Traversal):
            return self.build_traversal(source)
        if isinstance(source, NodeSet):
            if inspect.isclass(source.source) and issubclass(
                source.source, StructuredNode
            ):
                ident = self.build_label(source.source.__label__.lower(), source.source)
            else:
                ident = self.build_source(source.source)

            self.build_additional_match(ident, source)

            if hasattr(source, "order_by_elements"):
                self.build_order_by(ident, source)

            # source.filters seems to be used only by Traversal objects
            # source.q_filters is used by NodeSet objects
            if source.filters or source.q_filters:
                self.build_where_stmt(
                    ident=ident,
                    filters=source.filters,
                    source_class=source.source_class,
                    q_filters=source.q_filters,
                )

            return ident
        if isinstance(source, StructuredNode):
            return self.build_node(source)
        raise ValueError("Unknown source type " + repr(source))

    def create_relation_identifier(self) -> str:
        self._relation_identifier_count += 1
        return f"r{self._relation_identifier_count}"

    def create_node_identifier(self, prefix: str, path: str) -> str:
        if path not in self.node_set._unique_variables:
            self._node_identifier_count += 1
            return f"{prefix}{self._node_identifier_count}"
        return prefix

    def build_order_by(self, ident: str, source: "NodeSet") -> None:
        if "?" in source.order_by_elements:
            self._ast.with_clause = f"{ident}, rand() as r"
            self._ast.order_by = ["r"]
        else:
            order_by = []
            for elm in source.order_by_elements:
                if isinstance(elm, RawCypher):
                    order_by.append(elm.render({"n": ident}))
                    continue
                is_rel_property = "|" in elm
                if "__" not in elm and not is_rel_property:
                    prop = elm.split(" ")[0] if " " in elm else elm
                    if prop not in source.source_class.defined_properties(rels=False):
                        raise ValueError(
                            f"No such property {prop} on {source.source_class.__name__}. "
                            f"Note that Neo4j internals like id or element_id are not allowed "
                            f"for use in this operation."
                        )
                    order_by.append(f"{ident}.{elm}")
                else:
                    path, prop = elm.rsplit("__" if not is_rel_property else "|", 1)
                    result = self.lookup_query_variable(
                        path, return_relation=is_rel_property
                    )
                    if result:
                        order_by.append(f"{result[0]}.{prop}")
            self._ast.order_by = order_by

    def build_vector_query(self, vectorfilter: "VectorFilter", source: "NodeSet"):
        """
        Query a vector indexed property on the node. 
        """
        try:
            attribute = getattr(source, vectorfilter.vector_attribute_name)
        except AttributeError:
            raise # This raises the base AttributeError and provides potential correction

        if not attribute.vector_index:
            raise AttributeError(f"Attribute {vectorfilter.vector_attribute_name} is not declared with a vector index.")

        vectorfilter.index_name = f"vector_index_{source.__label__}_{vectorfilter.vector_attribute_name}"
        vectorfilter.nodeSetLabel = source.__label__.lower()

        self._ast.vector_index_query = vectorfilter
        self._ast.return_clause = f"{vectorfilter.nodeSetLabel}, score"
        self._ast.result_class = source.__class__
        

    def build_traversal(self, traversal: "Traversal") -> str:
        """
        traverse a relationship from a node to a set of nodes
        """
        # build source
        rhs_label = ":" + traversal.target_class.__label__

        # build source
        rel_ident = self.create_relation_identifier()
        lhs_ident = self.build_source(traversal.source)
        traversal_ident = f"{traversal.name}_{rel_ident}"
        rhs_ident = traversal_ident + rhs_label
        self._ast.return_clause = traversal_ident
        self._ast.result_class = traversal.target_class

        stmt = _rel_helper(
            lhs=lhs_ident,
            rhs=rhs_ident,
            ident=rel_ident,
            **traversal.definition,
        )
        self._ast.match.append(stmt)

        if traversal.filters:
            self.build_where_stmt(rel_ident, traversal.filters, traversal.source_class)

        return traversal_ident

    def _additional_return(self, name: str) -> None:
        if (
            not self._ast.additional_return or name not in self._ast.additional_return
        ) and name != self._ast.return_clause:
            if not self._ast.additional_return:
                self._ast.additional_return = []
            self._ast.additional_return.append(name)

    def build_traversal_from_path(
        self, relation: "Path", source_class: Any
    ) -> Tuple[str, Any]:
        path: str = relation.value
        stmt: str = ""
        source_class_iterator = source_class
        parts = re.split(path_split_regex, path)
        subgraph = self._ast.subgraph
        rel_iterator: str = ""
        already_present = False
        existing_rhs_name = ""
        for index, part in enumerate(parts):
            relationship = getattr(source_class_iterator, part)
            if rel_iterator:
                rel_iterator += "__"
            rel_iterator += part
            # build source
            if "node_class" not in relationship.definition:
                relationship.lookup_node_class()
            if not stmt:
                lhs_label = source_class_iterator.__label__
                lhs_name = lhs_label.lower()
                lhs_ident = f"{lhs_name}:{lhs_label}"
                if not index:
                    # This is the first one, we make sure that 'return'
                    # contains the primary node so _contains() works
                    # as usual
                    self._ast.return_clause = lhs_name
                    if self._subquery_namespace:
                        # Don't include label in identifier if we are in a subquery
                        lhs_ident = lhs_name
                elif relation.include_nodes_in_return:
                    self._additional_return(lhs_name)
            else:
                lhs_ident = stmt

            already_present = part in subgraph
            rel_ident = self.create_relation_identifier()
            rhs_label = relationship.definition["node_class"].__label__
            if relation.relation_filtering:
                rhs_name = rel_ident
                rhs_ident = f":{rhs_label}"
            else:
                if index + 1 == len(parts) and relation.alias:
                    # If an alias is defined, use it to store the last hop in the path
                    rhs_name = relation.alias
                else:
                    rhs_name = f"{rhs_label.lower()}_{rel_iterator}"
                    rhs_name = self.create_node_identifier(rhs_name, rel_iterator)
                rhs_ident = f"{rhs_name}:{rhs_label}"

            if relation.include_nodes_in_return and not already_present:
                self._additional_return(rhs_name)

            if not already_present:
                subgraph[part] = {
                    "target": relationship.definition["node_class"],
                    "children": {},
                    "variable_name": rhs_name,
                    "rel_variable_name": rel_ident,
                }
            else:
                existing_rhs_name = subgraph[part][
                    (
                        "rel_variable_name"
                        if relation.relation_filtering
                        else "variable_name"
                    )
                ]
            if relation.include_rels_in_return and not already_present:
                self._additional_return(rel_ident)
            stmt = _rel_helper(
                lhs=lhs_ident,
                rhs=rhs_ident,
                ident=rel_ident,
                direction=relationship.definition["direction"],
                relation_type=relationship.definition["relation_type"],
            )
            source_class_iterator = relationship.definition["node_class"]
            subgraph = subgraph[part]["children"]

        if not already_present:
            if relation.optional:
                self._ast.optional_match.append(stmt)
            else:
                self._ast.match.append(stmt)
            return rhs_name, relationship.definition["node_class"]

        return existing_rhs_name, relationship.definition["node_class"]

    def build_node(self, node: StructuredNode) -> str:
        ident = node.__class__.__name__.lower()
        place_holder = self._register_place_holder(ident)

        # Hack to emulate START to lookup a node by id
        _node_lookup = f"MATCH ({ident}) WHERE {db.get_id_method()}({ident})=${place_holder} WITH {ident}"
        self._ast.lookup = _node_lookup

        self._query_params[place_holder] = db.parse_element_id(node.element_id)

        self._ast.return_clause = ident
        self._ast.result_class = node.__class__
        return ident

    def build_label(self, ident: str, cls: type[StructuredNode]) -> str:
        """
        match nodes by a label
        """
        ident_w_label = ident + ":" + cls.__label__

        if not self._ast.return_clause:
            if (
                not self._ast.additional_return
                or ident not in self._ast.additional_return
            ):
                self._ast.match.append(f"({ident_w_label})")
                self._ast.return_clause = ident
                self._ast.result_class = cls
        elif not self._ast.match:
            # If we get here, it means return_clause was filled because of an
            # optional match, so we add a regular match for root node.
            # Not very elegant, this part would deserve a refactoring...
            self._ast.match.append(f"({ident_w_label})")
            self._ast.result_class = cls
        return ident

    def build_additional_match(self, ident: str, node_set: "NodeSet") -> None:
        """
        handle additional matches supplied by 'has()' calls
        """
        source_ident = ident

        for _, value in node_set.must_match.items():
            if isinstance(value, dict):
                label = ":" + value["node_class"].__label__
                stmt = f"EXISTS ({_rel_helper(lhs=source_ident, rhs=label, ident='', **value)})"
                self._ast.where.append(stmt)
            else:
                raise ValueError("Expecting dict got: " + repr(value))

        for _, val in node_set.dont_match.items():
            if isinstance(val, dict):
                label = ":" + val["node_class"].__label__
                stmt = f"NOT EXISTS ({_rel_helper(lhs=source_ident, rhs=label, ident='', **val)})"
                self._ast.where.append(stmt)
            else:
                raise ValueError("Expecting dict got: " + repr(val))

    def _register_place_holder(self, key: str) -> str:
        if key in self._place_holder_registry:
            self._place_holder_registry[key] += 1
        else:
            self._place_holder_registry[key] = 1
        place_holder = f"{key}_{self._place_holder_registry[key]}"
        if self._subquery_namespace:
            place_holder = f"{self._subquery_namespace}_{place_holder}"
        return place_holder

    def _parse_path(
        self, source_class: type[StructuredNode], prop: str
    ) -> Tuple[str, str, str, Any, bool]:
        is_rel_filter = "|" in prop
        if is_rel_filter:
            path, prop = prop.rsplit("|", 1)
        else:
            path, prop = prop.rsplit("__", 1)
        result = self.lookup_query_variable(path, return_relation=is_rel_filter)
        is_optional_relation = False
        if not result:
            ident, target_class = self.build_traversal_from_path(
                Path(
                    value=path,
                    relation_filtering=is_rel_filter,
                ),
                source_class,
            )
        else:
            ident, target_class, is_optional_relation = result
        return ident, path, prop, target_class, is_optional_relation

    def _finalize_filter_statement(
        self, operator: str, ident: str, prop: str, val: Any
    ) -> str:
        if operator in _UNARY_OPERATORS:
            # unary operators do not have a parameter
            statement = f"{ident}.{prop} {operator}"
        else:
            place_holder = self._register_place_holder(ident + "_" + prop)
            if operator == _SPECIAL_OPERATOR_ARRAY_IN:
                statement = operator.format(
                    ident=ident,
                    prop=prop,
                    val=f"${place_holder}",
                )
            else:
                statement = f"{ident}.{prop} {operator} ${place_holder}"
            self._query_params[place_holder] = val

        return statement

    def _build_filter_statements(
        self,
        ident: str,
        filters: dict[str, tuple],
        target: list[tuple[str, bool]],
        source_class: type[StructuredNode],
    ) -> None:
        for prop, op_and_val in filters.items():
            path = None
            is_rel_filter = "|" in prop
            target_class = source_class
            is_optional_relation = False
            if "__" in prop or is_rel_filter:
                (
                    ident,
                    path,
                    prop,
                    target_class,
                    is_optional_relation,
                ) = self._parse_path(source_class, prop)
            operator, val = op_and_val
            if not is_rel_filter:
                prop = target_class.defined_properties(rels=False)[
                    prop
                ].get_db_property_name(prop)
            statement = self._finalize_filter_statement(operator, ident, prop, val)
            target.append((statement, is_optional_relation))

    def _parse_q_filters(
        self, ident: str, q: Union[QBase, Any], source_class: type[StructuredNode]
    ) -> tuple[str, str]:
        target: list[tuple[str, bool]] = []

        def add_to_target(statement: str, connector: Q, optional: bool) -> None:
            if not statement:
                return
            if connector == Q.OR:
                statement = f"({statement})"
            target.append((statement, optional))

        for child in q.children:
            if isinstance(child, QBase):
                q_childs, q_opt_childs = self._parse_q_filters(
                    ident, child, source_class
                )
                add_to_target(q_childs, child.connector, False)
                add_to_target(q_opt_childs, child.connector, True)
            else:
                kwargs = {child[0]: child[1]}
                filters = process_filter_args(source_class, kwargs)
                self._build_filter_statements(ident, filters, target, source_class)
        match_filters = [filter[0] for filter in target if not filter[1]]
        opt_match_filters = [filter[0] for filter in target if filter[1]]
        if q.connector == Q.OR and match_filters and opt_match_filters:
            # In this case, we can't split filters in two WHERE statements so we move
            # everything into the one applied after OPTIONAL MATCH statements...
            opt_match_filters += match_filters
            match_filters = []
            self._ast.mixed_filters = True

        ret = f" {q.connector} ".join(match_filters)
        if ret and q.negated:
            ret = f"NOT ({ret})"
        opt_ret = f" {q.connector} ".join(opt_match_filters)
        if opt_ret and q.negated:
            opt_ret = f"NOT ({opt_ret})"
        return ret, opt_ret

    def build_where_stmt(
        self,
        ident: str,
        filters: list,
        source_class: type[StructuredNode],
        q_filters: Union[QBase, Any, None] = None,
    ) -> None:
        """
        Construct a where statement from some filters.

        We make a difference between filters applied to variables coming from MATCH and
        OPTIONAL MATCH statements.

        """
        if q_filters is not None:
            stmt, opt_stmt = self._parse_q_filters(ident, q_filters, source_class)
            if stmt:
                self._ast.where.append(stmt)
            if opt_stmt:
                self._ast.optional_where.append(opt_stmt)
        else:
            stmts = []
            for row in filters:
                negate = False

                # pre-process NOT cases as they are nested dicts
                if "__NOT__" in row and len(row) == 1:
                    negate = True
                    row = row["__NOT__"]

                for prop, operator_and_val in row.items():
                    operator, val = operator_and_val
                    if operator in _UNARY_OPERATORS:
                        # unary operators do not have a parameter
                        statement = (
                            f"{'NOT' if negate else ''} {ident}.{prop} {operator}"
                        )
                    else:
                        place_holder = self._register_place_holder(ident + "_" + prop)
                        statement = f"{'NOT' if negate else ''} {ident}.{prop} {operator} ${place_holder}"
                        self._query_params[place_holder] = val
                    stmts.append(statement)

            self._ast.where.append(" AND ".join(stmts))

    def lookup_query_variable(
        self, path: str, return_relation: bool = False
    ) -> TOptional[Tuple[str, Any, bool]]:
        """Retrieve the variable name generated internally for the given traversal path."""
        subgraph = self._ast.subgraph
        if not subgraph:
            return None
        traversals = re.split(path_split_regex, path)
        if len(traversals) == 0:
            raise ValueError("Can only lookup traversal variables")
        if traversals[0] not in subgraph:
            return None

        # Check if relation is coming from an optional MATCH
        # (declared using fetch|traverse_relations)
        is_optional_relation = False
        for relation in self.node_set.relations_to_fetch:
            if relation.value == path:
                is_optional_relation = relation.optional
                break

        subgraph = subgraph[traversals[0]]
        if len(traversals) == 1:
            variable_to_return = f"{subgraph['rel_variable_name' if return_relation else 'variable_name']}"
            return variable_to_return, subgraph["target"], is_optional_relation
        variable_to_return = ""
        last_property = traversals[-1]
        for part in traversals[1:]:
            child = subgraph["children"].get(part)
            if not child:
                return None
            subgraph = child
            if part == last_property:
                # if last part of prop is the last traversal
                # we are safe to lookup the variable from the query
                variable_to_return = f"{subgraph['rel_variable_name' if return_relation else 'variable_name']}"
        return variable_to_return, subgraph["target"], is_optional_relation

    def build_query(self) -> str:
        query: str = ""

        if self._ast.lookup:
            query += self._ast.lookup

        if self._ast.vector_index_query:
            query += f"""CALL () {{ 
                CALL db.index.vector.queryNodes("{self._ast.vector_index_query.index_name}", {self._ast.vector_index_query.topk}, {self._ast.vector_index_query.vector}) 
                YIELD node AS {self._ast.vector_index_query.nodeSetLabel}, score 
                RETURN {self._ast.vector_index_query.nodeSetLabel}, score 
                }}"""
            # This ensures that we bring the context of the new nodeSet and score along with us for further filtering
            query += f""" WITH {self._ast.vector_index_query.nodeSetLabel}, score"""

        # Instead of using only one MATCH statement for every relation
        # to follow, we use one MATCH per relation (to avoid cartesian
        # product issues...).
        # There might be optimizations to be done, using projections,
        # or pusing patterns instead of a chain of OPTIONAL MATCH.
        if self._ast.match:
            query += " MATCH "
            query += " MATCH ".join(i for i in self._ast.match)

        if self._ast.where:
            query += " WHERE "
            query += " AND ".join(self._ast.where)

        if self._ast.optional_match:
            query += " OPTIONAL MATCH "
            query += " OPTIONAL MATCH ".join(i for i in self._ast.optional_match)

        if self._ast.optional_where:
            if self._ast.mixed_filters:
                # Make sure filtering works as expected with optional match, even if it's not performant...
                query += " WITH *"
            query += " WHERE "
            query += " AND ".join(self._ast.optional_where)

        if self._ast.with_clause:
            query += " WITH "
            query += self._ast.with_clause

        returned_items: list[str] = []
        if hasattr(self.node_set, "_subqueries"):
            for subquery in self.node_set._subqueries:
                query += " CALL {"
                if subquery["initial_context"]:
                    query += " WITH "
                    context: list[str] = []
                    for var in subquery["initial_context"]:
                        if isinstance(var, (NodeNameResolver, RelationNameResolver)):
                            context.append(var.resolve(self))
                        else:
                            context.append(var)
                    query += ",".join(context)

                query += f"{subquery['query']} }} "
                self._query_params.update(subquery["query_params"])
                for varname in subquery["return_set"]:
                    # We declare the returned variables as "virtual" relations of the
                    # root node class to make sure they will be translated by a call to
                    # resolve_subgraph() (otherwise, they will be lost).
                    # This is probably a temporary solution until we find something better...
                    self._ast.subgraph[varname] = {
                        "target": None,  # We don't need target class in this use case
                        "children": {},
                        "variable_name": varname,
                        "rel_variable_name": varname,
                    }
                returned_items += subquery["return_set"]

        if hasattr(self.node_set, "_intermediate_transforms"):
            for transform in self.node_set._intermediate_transforms:
                query += " WITH "
                query += "DISTINCT " if transform.get("distinct") else ""
                injected_vars: list = []
                # Reset return list since we'll probably invalidate most variables
                self._ast.return_clause = ""
                self._ast.additional_return = []
                returned_items = []
                for name, varprops in transform["vars"].items():
                    source = varprops["source"]
                    if isinstance(source, (NodeNameResolver, RelationNameResolver)):
                        transformation = source.resolve(self)
                    else:
                        transformation = source
                    if varprops.get("source_prop"):
                        transformation += f".{varprops['source_prop']}"
                    transformation += f" AS {name}"
                    if varprops.get("include_in_return"):
                        returned_items += [name]
                    injected_vars.append(transformation)
                query += ",".join(injected_vars)
                if not transform["ordering"]:
                    continue
                query += " ORDER BY "
                ordering: list = []
                for item in transform["ordering"]:
                    if isinstance(item, RawCypher):
                        ordering.append(item.render({}))
                        continue
                    if item.startswith("-"):
                        ordering.append(f"{item[1:]} DESC")
                    else:
                        ordering.append(item)
                query += ",".join(ordering)

        query += " RETURN "
        if self._ast.return_clause and not self._subquery_namespace:
            returned_items.append(self._ast.return_clause)
        if self._ast.additional_return:
            returned_items += self._ast.additional_return
        if hasattr(self.node_set, "_extra_results"):
            for props in self.node_set._extra_results:
                leftpart = props["vardef"].render(self)
                varname = (
                    props["alias"]
                    if props.get("alias")
                    else props["vardef"].get_internal_name()
                )
                if varname in returned_items:
                    # We're about to override an existing variable, delete it first to
                    # avoid duplicate error
                    returned_items.remove(varname)
                returned_items.append(f"{leftpart} AS {varname}")

        query += ", ".join(returned_items)

        if self._ast.order_by:
            query += " ORDER BY "
            query += ", ".join(self._ast.order_by)

        # If we return a count with pagination, pagination has to happen before RETURN
        # It will then be included in the WITH clause already
        if self._ast.skip and not self._ast.is_count:
            query += f" SKIP {self._ast.skip}"

        if self._ast.limit and not self._ast.is_count:
            query += f" LIMIT {self._ast.limit}"
        print(query)
        return query

    def _count(self) -> int:
        self._ast.is_count = True
        # If we return a count with pagination, pagination has to happen before RETURN
        # Like : WITH my_var SKIP 10 LIMIT 10 RETURN count(my_var)
        self._ast.with_clause = f"{self._ast.return_clause}"
        if self._ast.skip:
            self._ast.with_clause += f" SKIP {self._ast.skip}"

        if self._ast.limit:
            self._ast.with_clause += f" LIMIT {self._ast.limit}"

        self._ast.return_clause = f"count({self._ast.return_clause})"
        # drop order_by, results in an invalid query
        self._ast.order_by = None
        # drop additional_return to avoid unexpected result
        self._ast.additional_return = None
        query = self.build_query()
        results, _ = db.cypher_query(query, self._query_params)
        return int(results[0][0])

    def _contains(self, node_element_id: TOptional[Union[str, int]]) -> bool:
        # inject id = into ast
        if not self._ast.return_clause and self._ast.additional_return:
            self._ast.return_clause = self._ast.additional_return[0]
        if not self._ast.return_clause:
            raise ValueError("Cannot use contains without a return clause")
        ident = self._ast.return_clause
        place_holder = self._register_place_holder(ident + "_contains")
        self._ast.where.append(f"{db.get_id_method()}({ident}) = ${place_holder}")
        self._query_params[place_holder] = node_element_id
        return self._count() >= 1

    def _execute(self, lazy: bool = False, dict_output: bool = False) -> Any:
        if lazy:
            # inject id() into return or return_set
            if self._ast.return_clause:
                self._ast.return_clause = (
                    f"{db.get_id_method()}({self._ast.return_clause})"
                )
            else:
                if self._ast.additional_return is not None:
                    self._ast.additional_return = [
                        f"{db.get_id_method()}({item})"
                        for item in self._ast.additional_return
                    ]
        query = self.build_query()
        results, prop_names = db.cypher_query(
            query,
            self._query_params,
            resolve_objects=True,
        )
        if dict_output:
            for item in results:
                yield dict(zip(prop_names, item))
            return
        # The following is not as elegant as it could be but had to be copied from the
        # version prior to cypher_query with the resolve_objects capability.
        # It seems that certain calls are only supposed to be focusing to the first
        # result item returned (?)
        if results and len(results[0]) == 1:
            for n in results:
                yield n[0]
        else:
            for result in results:
                yield result


class BaseSet:
    """
    Base class for all node sets.

    Contains common python magic methods, __len__, __contains__ etc
    """

    query_cls = QueryBuilder
    source_class: type[StructuredNode]

    def all(self, lazy: bool = False) -> list:
        """
        Return all nodes belonging to the set
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :return: list of nodes
        :rtype: list
        """
        ast = self.query_cls(self).build_ast()
        results = [
            node for node in ast._execute(lazy)
        ]  # Collect all nodes asynchronously
        return results

    def __iter__(self) -> Iterator:
        ast = self.query_cls(self).build_ast()
        for item in ast._execute():
            yield item

    def __len__(self) -> int:
        ast = self.query_cls(self).build_ast()
        return ast._count()

    def __bool__(self) -> bool:
        """
        Override for __bool__ dunder method.
        :return: True if the set contains any nodes, False otherwise
        :rtype: bool
        """
        ast = self.query_cls(self).build_ast()
        _count = ast._count()
        return _count > 0

    def __nonzero__(self) -> bool:
        """
        Override for __bool__ dunder method.
        :return: True if the set contains any node, False otherwise
        :rtype: bool
        """
        return self.__bool__()

    def __contains__(self, obj: Union[StructuredNode, Any]) -> bool:
        if isinstance(obj, StructuredNode):
            if hasattr(obj, "element_id") and obj.element_id is not None:
                ast = self.query_cls(self).build_ast()
                obj_element_id = db.parse_element_id(obj.element_id)
                return ast._contains(obj_element_id)
            raise ValueError("Unsaved node: " + repr(obj))

        raise ValueError("Expecting StructuredNode instance")

    def __getitem__(self, key: Union[int, slice]) -> TOptional["BaseSet"]:
        if isinstance(key, slice):
            if key.stop and key.start:
                self.limit = key.stop - key.start
                self.skip = key.start
            elif key.stop:
                self.limit = key.stop
            elif key.start:
                self.skip = key.start

            return self

        if isinstance(key, int):
            self.skip = key
            self.limit = 1

            ast = self.query_cls(self).build_ast()
            _first_item = [node for node in ast._execute()][0]
            return _first_item

        return None


@dataclass
class Optional:  # type: ignore[no-redef]
    """Simple relation qualifier."""

    relation: str


@dataclass
class Path:
    """Path traversal definition."""

    value: str
    optional: bool = False
    include_nodes_in_return: bool = True
    include_rels_in_return: bool = True
    relation_filtering: bool = False
    alias: TOptional[str] = None


@dataclass
class RelationNameResolver:
    """Helper to refer to a relation variable name.

    Since variable names are generated automatically within MATCH statements (for
    anything injected using fetch_relations or traverse_relations), we need a way to
    retrieve them.

    """

    relation: str

    def resolve(self, qbuilder: QueryBuilder) -> str:
        result = qbuilder.lookup_query_variable(self.relation, True)
        if result is None:
            raise ValueError(
                f"Unable to resolve variable name for relation {self.relation}"
            )
        return result[0]


@dataclass
class NodeNameResolver:
    """Helper to refer to a node variable name.

    Since variable names are generated automatically within MATCH statements (for
    anything injected using fetch_relations or traverse_relations), we need a way to
    retrieve them.

    """

    node: str

    def resolve(self, qbuilder: QueryBuilder) -> str:
        if self.node == "self" and qbuilder._ast.return_clause:
            return qbuilder._ast.return_clause
        result = qbuilder.lookup_query_variable(self.node)
        if result is None:
            raise ValueError(f"Unable to resolve variable name for node {self.node}")
        return result[0]


@dataclass
class BaseFunction:
    input_name: Union[str, "BaseFunction", NodeNameResolver, RelationNameResolver]

    def __post_init__(self) -> None:
        self._internal_name: str = ""

    def get_internal_name(self) -> str:
        return self._internal_name

    def resolve_internal_name(self, qbuilder: QueryBuilder) -> str:
        if isinstance(self.input_name, (NodeNameResolver, RelationNameResolver)):
            self._internal_name = self.input_name.resolve(qbuilder)
        else:
            self._internal_name = str(self.input_name)
        return self._internal_name

    def render(self, qbuilder: QueryBuilder) -> str:
        raise NotImplementedError


@dataclass
class AggregatingFunction(BaseFunction):
    """Base aggregating function class."""

    pass


@dataclass
class Collect(AggregatingFunction):
    """collect() function."""

    distinct: bool = False

    def render(self, qbuilder: QueryBuilder) -> str:
        varname = self.resolve_internal_name(qbuilder)
        if self.distinct:
            return f"collect(DISTINCT {varname})"
        return f"collect({varname})"


@dataclass
class ScalarFunction(BaseFunction):
    """Base scalar function class."""

    @property
    def function_name(self) -> str:
        raise NotImplementedError

    def render(self, qbuilder: QueryBuilder) -> str:
        if isinstance(self.input_name, str):
            content = str(self.input_name)
        elif isinstance(self.input_name, BaseFunction):
            content = self.input_name.render(qbuilder)
            self._internal_name = self.input_name.get_internal_name()
        else:
            content = self.resolve_internal_name(qbuilder)
        return f"{self.function_name}({content})"


@dataclass
class Last(ScalarFunction):
    """last() function."""

    @property
    def function_name(self) -> str:
        return "last"


@dataclass
class Size(ScalarFunction):
    """size() function."""

    @property
    def function_name(self) -> str:
        return "size"


@dataclass
class RawCypher:
    """Helper to inject raw cypher statement.

    Can be used in order_by() call for example.

    """

    statement: str

    def __post_init__(self) -> None:
        if CYPHER_ACTIONS_WITH_SIDE_EFFECT_EXPR.search(self.statement):
            raise ValueError(
                "RawCypher: Do not include any action that has side effect"
            )

    def render(self, context: dict) -> str:
        return string.Template(self.statement).substitute(context)


class NodeSet(BaseSet):
    """
    A class representing as set of nodes matching common query parameters
    """

    def __init__(self, source: Any) -> None:
        self.source = source  # could be a Traverse object or a node class
        if isinstance(source, Traversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, StructuredNode):
            self.source_class = source
        elif isinstance(source, StructuredNode):
            self.source_class = source.__class__
        else:
            raise ValueError("Bad source for nodeset " + repr(source))

        # setup Traversal objects using relationship definitions
        install_traversals(self.source_class, self)

        self.filters: list = []
        self.q_filters = Q()
        self.order_by_elements: list = []

        # used by has()
        self.must_match: dict = {}
        self.dont_match: dict = {}

        self.relations_to_fetch: list[Path] = []
        self._extra_results: list = []
        self._subqueries: list[Subquery] = []
        self._intermediate_transforms: list = []
        self._unique_variables: list[str] = []
        self._vector_query: str = None 

    def __await__(self) -> Any:
        return self.all().__await__()  # type: ignore[attr-defined]

    def _get(
        self, limit: TOptional[int] = None, lazy: bool = False, **kwargs: dict[str, Any]
    ) -> list:
        self.filter(**kwargs)
        if limit:
            self.limit = limit
        ast = self.query_cls(self).build_ast()
        results = [node for node in ast._execute(lazy)]
        return results

    def get(self, lazy: bool = False, **kwargs: Any) -> Any:
        """
        Retrieve one node from the set matching supplied parameters
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = self._get(limit=2, lazy=lazy, **kwargs)
        if len(result) > 1:
            raise MultipleNodesReturned(repr(kwargs))
        if not result:
            raise self.source_class.DoesNotExist(repr(kwargs))
        return result[0]

    def get_or_none(self, **kwargs: Any) -> Any:
        """
        Retrieve a node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return self.get(**kwargs)
        except self.source_class.DoesNotExist:
            return None

    def first(self, **kwargs: Any) -> Any:
        """
        Retrieve the first node from the set matching supplied parameters

        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = self._get(limit=1, **kwargs)
        if result:
            return result[0]
        else:
            raise self.source_class.DoesNotExist(repr(kwargs))

    def first_or_none(self, **kwargs: Any) -> Any:
        """
        Retrieve the first node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return self.first(**kwargs)
        except self.source_class.DoesNotExist:
            pass
        return None

    def filter(self, *args: Any, **kwargs: Any) -> "BaseSet":
        """
        Apply filters to the existing nodes in the set.

        :param args: a Q object

            e.g `.filter(Q(salary__lt=10000) | Q(salary__gt=20000))`.

        :param kwargs: filter parameters

            Filters mimic Django's syntax with the double '__' to separate field and operators.

            e.g `.filter(salary__gt=20000)` results in `salary > 20000`.

            The following operators are available:

             * 'lt': less than
             * 'gt': greater than
             * 'lte': less than or equal to
             * 'gte': greater than or equal to
             * 'ne': not equal to
             * 'in': matches one of list (or tuple)
             * 'isnull': is null
             * 'regex': matches supplied regex (neo4j regex format)
             * 'exact': exactly match string (just '=')
             * 'iexact': case insensitive match string
             * 'contains': contains string
             * 'icontains': case insensitive contains
             * 'startswith': string starts with
             * 'istartswith': case insensitive string starts with
             * 'endswith': string ends with
             * 'iendswith': case insensitive string ends with

        :return: self
        """
        if args or kwargs:
            # Need to grab and remove the VectorFilter from both args and kwargs
            new_args = [] # As args are a tuple, theyre immutable. But we need to remove the vectorfilter from the arguments so they dont go into Q. 
            for arg in args:
                if isinstance(arg, VectorFilter) and (not self._vector_query):
                    self._vector_query = arg
                new_args.append(arg)

            new_args = tuple(new_args)

            if kwargs.get("vector_filter"):
                if isinstance(kwargs["vector_filter"], VectorFilter) and (not self._vector_query):
                    self._vector_query = kwargs.pop("vector_filter")
                    

            self.q_filters = Q(self.q_filters & Q(*new_args, **kwargs))

        return self

    def exclude(self, *args: Any, **kwargs: Any) -> "BaseSet":
        """
        Exclude nodes from the NodeSet via filters.

        :param kwargs: filter parameters see syntax for the filter method
        :return: self
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & ~Q(*args, **kwargs))
        return self

    def has(self, **kwargs: Any) -> "BaseSet":
        must_match, dont_match = process_has_args(self.source_class, kwargs)
        self.must_match.update(must_match)
        self.dont_match.update(dont_match)
        return self

    def order_by(self, *props: Any) -> "BaseSet":
        """
        Order by properties. Prepend with minus to do descending. Pass None to
        remove ordering.
        """
        should_remove = len(props) == 1 and props[0] is None
        if not hasattr(self, "order_by_elements") or should_remove:
            self.order_by_elements = []
            if should_remove:
                return self
        if "?" in props:
            self.order_by_elements.append("?")
        else:
            for prop in props:
                if isinstance(prop, RawCypher):
                    self.order_by_elements.append(prop)
                    continue
                prop = prop.strip()
                if prop.startswith("-"):
                    prop = prop[1:]
                    desc = True
                else:
                    desc = False

                if prop in self.source_class.defined_properties(rels=False):
                    property_obj = getattr(self.source_class, prop)
                    if isinstance(property_obj, AliasProperty):
                        prop = property_obj.aliased_to()

                self.order_by_elements.append(prop + (" DESC" if desc else ""))

        return self

    def _register_relation_to_fetch(
        self, relation_def: Any, alias: TOptional[str] = None
    ) -> "Path":
        if isinstance(relation_def, Path):
            item = relation_def
        else:
            item = Path(
                value=relation_def,
            )
        if alias:
            item.alias = alias
        return item

    def unique_variables(self, *paths: tuple[str, ...]) -> "NodeSet":
        """Generate unique variable names for the given paths."""
        self._unique_variables = paths
        return self

    def traverse(self, *paths: tuple[str, ...], **aliased_paths: dict) -> "NodeSet":
        """Specify a set of paths to traverse."""
        relations = []
        for path in paths:
            relations.append(self._register_relation_to_fetch(path))
        for alias, aliased_path in aliased_paths.items():
            relations.append(
                self._register_relation_to_fetch(aliased_path, alias=alias)
            )
        self.relations_to_fetch = relations
        return self

    def fetch_relations(self, *relation_names: tuple[str, ...]) -> "NodeSet":
        """Specify a set of relations to traverse and return."""
        warnings.warn(
            "fetch_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )
        relations = []
        for relation_name in relation_names:
            if isinstance(relation_name, Optional):
                relation_name = Path(value=relation_name.relation, optional=True)
            relations.append(self._register_relation_to_fetch(relation_name))
        self.relations_to_fetch = relations
        return self

    def traverse_relations(
        self, *relation_names: tuple[str, ...], **aliased_relation_names: dict
    ) -> "NodeSet":
        """Specify a set of relations to traverse only."""

        warnings.warn(
            "traverse_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )

        def convert_to_path(input: Union[str, Optional]) -> Path:
            if isinstance(input, Optional):
                path = Path(value=input.relation, optional=True)
            else:
                path = Path(value=input)
            path.include_nodes_in_return = False
            path.include_rels_in_return = False
            return path

        relations = []
        for relation_name in relation_names:
            relations.append(
                self._register_relation_to_fetch(convert_to_path(relation_name))
            )
        for alias, relation_def in aliased_relation_names.items():
            relations.append(
                self._register_relation_to_fetch(
                    convert_to_path(relation_def), alias=alias
                )
            )

        self.relations_to_fetch = relations
        return self

    def annotate(self, *vars: tuple, **aliased_vars: tuple) -> "NodeSet":
        """Annotate node set results with extra variables."""

        def register_extra_var(
            vardef: Union[AggregatingFunction, ScalarFunction, Any],
            varname: Union[str, None] = None,
        ) -> None:
            if isinstance(vardef, (AggregatingFunction, ScalarFunction)):
                self._extra_results.append(
                    {"vardef": vardef, "alias": varname if varname else ""}
                )
            else:
                raise NotImplementedError

        for vardef in vars:
            register_extra_var(vardef)
        for varname, vardef in aliased_vars.items():
            register_extra_var(vardef, varname)

        return self

    def _to_subgraph(self, root_node: Any, other_nodes: Any, subgraph: dict) -> Any:
        """Recursive method to build root_node's relation graph from subgraph."""
        root_node._relations = {}
        for name, relation_def in subgraph.items():
            for var_name, node in other_nodes.items():
                if (
                    var_name
                    not in [
                        relation_def["variable_name"],
                        relation_def["rel_variable_name"],
                    ]
                    or node is None
                ):
                    continue
                if isinstance(node, list):
                    if len(node) > 0 and isinstance(node[0], StructuredRel):
                        name += "_relationship"
                    root_node._relations[name] = []
                    for item in node:
                        root_node._relations[name].append(
                            self._to_subgraph(
                                item, other_nodes, relation_def["children"]
                            )
                        )
                else:
                    if isinstance(node, StructuredRel):
                        name += "_relationship"
                    root_node._relations[name] = self._to_subgraph(
                        node, other_nodes, relation_def["children"]
                    )

        return root_node

    def resolve_subgraph(self) -> list:
        """
        Convert every result contained in this node set to a subgraph.

        By default, we receive results from neomodel as a list of
        nodes without the hierarchy. This method tries to rebuild this
        hierarchy without overriding anything in the node, that's why
        we use a dedicated property to store node's relations.

        """
        if (
            self.relations_to_fetch
            and not self.relations_to_fetch[0].include_nodes_in_return
            and not self.relations_to_fetch[0].include_rels_in_return
        ):
            raise NotImplementedError(
                "You cannot use traverse_relations() with resolve_subgraph(), use fetch_relations() instead."
            )
        results: list = []
        qbuilder = self.query_cls(self)
        qbuilder.build_ast()
        if not qbuilder._ast.subgraph:
            raise RuntimeError(
                "Nothing to resolve. Make sure to include relations in the result using fetch_relations() or filter()."
            )
        other_nodes = {}
        root_node = None
        for row in qbuilder._execute(dict_output=True):
            for name, node in row.items():
                if node.__class__ is self.source and "_" not in name:
                    root_node = node
                    continue
                if isinstance(node, list) and isinstance(node[0], list):
                    other_nodes[name] = node[0]
                    continue
                other_nodes[name] = node
            results.append(
                self._to_subgraph(root_node, other_nodes, qbuilder._ast.subgraph)
            )
        return results

    def subquery(
        self,
        nodeset: "NodeSet",
        return_set: list[str],
        initial_context: TOptional[list[str]] = None,
    ) -> "NodeSet":
        """Add a subquery to this node set.

        A subquery is a regular cypher query but executed within the context of a CALL
        statement. Such query will generally fetch additional variables which must be
        declared inside return_set variable in order to be included in the final RETURN
        statement.
        """
        namespace = f"sq{len(self._subqueries) + 1}"
        qbuilder = nodeset.query_cls(nodeset, subquery_namespace=namespace).build_ast()
        for var in return_set:
            if (
                var != qbuilder._ast.return_clause
                and (
                    not qbuilder._ast.additional_return
                    or var not in qbuilder._ast.additional_return
                )
                and var
                not in [res["alias"] for res in nodeset._extra_results if res["alias"]]
                and var
                not in [
                    varname
                    for tr in nodeset._intermediate_transforms
                    for varname, vardef in tr["vars"].items()
                    if vardef.get("include_in_return")
                ]
            ):
                raise RuntimeError(f"Variable '{var}' is not returned by subquery.")
        if initial_context:
            for var in initial_context:
                if type(var) is not str and not isinstance(
                    var, (NodeNameResolver, RelationNameResolver, RawCypher)
                ):
                    raise ValueError(
                        f"Wrong variable specified in initial context, should be a string or an instance of NodeNameResolver or RelationNameResolver"
                    )
        self._subqueries.append(
            {
                "query": qbuilder.build_query(),
                "query_params": qbuilder._query_params,
                "return_set": return_set,
                "initial_context": initial_context,
            }
        )
        return self

    def intermediate_transform(
        self,
        vars: dict[str, Transformation],
        distinct: bool = False,
        ordering: TOptional[list] = None,
    ) -> "NodeSet":
        if not vars:
            raise ValueError(
                "You must provide one variable at least when calling intermediate_transform()"
            )
        for name, props in vars.items():
            source = props["source"]
            if type(source) is not str and not isinstance(
                source, (NodeNameResolver, RelationNameResolver, RawCypher)
            ):
                raise ValueError(
                    f"Wrong source type specified for variable '{name}', should be a string or an instance of NodeNameResolver or RelationNameResolver"
                )
        self._intermediate_transforms.append(
            {"vars": vars, "distinct": distinct, "ordering": ordering}
        )
        return self


class Traversal(BaseSet):
    """
    Models a traversal from a node to another.

    :param source: Starting of the traversal.
    :type source: A :class:`~neomodel.core.StructuredNode` subclass, an
                  instance of such, a :class:`~neomodel.match.NodeSet` instance
                  or a :class:`~neomodel.match.Traversal` instance.
    :param name: A name for the traversal.
    :type name: :class:`str`
    :param definition: A relationship definition that most certainly deserves
                       a documentation here.
    :type definition: :class:`dict`
    """

    definition: dict
    source: Any
    source_class: Any
    target_class: Any
    name: str
    filters: list

    def __await__(self) -> Any:
        return self.all().__await__()  # type: ignore[attr-defined]

    def __init__(self, source: Any, name: str, definition: dict) -> None:
        """
        Create a traversal

        """
        self.source = source

        if isinstance(source, Traversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, StructuredNode):
            self.source_class = source
        elif isinstance(source, StructuredNode):
            self.source_class = source.__class__
        elif isinstance(source, NodeSet):
            self.source_class = source.source_class
        else:
            raise TypeError(f"Bad source for traversal: {type(source)}")

        invalid_keys = set(definition) - {
            "direction",
            "model",
            "node_class",
            "relation_type",
        }
        if invalid_keys:
            raise ValueError(f"Prohibited keys in Traversal definition: {invalid_keys}")

        self.definition = definition
        self.target_class = definition["node_class"]
        self.name = name
        self.filters: list = []

    def match(self, **kwargs: Any) -> "Traversal":
        """
        Traverse relationships with properties matching the given parameters.

            e.g: `.match(price__lt=10)`

        :param kwargs: see `NodeSet.filter()` for syntax
        :return: self
        """
        if kwargs:
            if self.definition.get("model") is None:
                raise ValueError(
                    "match() with filter only available on relationships with a model"
                )
            output = process_filter_args(self.definition["model"], kwargs)
            if output:
                self.filters.append(output)
        return self



from neomodel import (
    ArrayProperty,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    BooleanProperty,
    DateProperty,
    DateTimeFormatProperty,
    DateTimeProperty,
    FloatProperty,
    IntegerProperty,
    RelationshipFrom,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    UniqueIdProperty,
)
from neomodel.contrib.spatial_properties import PointProperty


class Document(StructuredNode):
    uid = UniqueIdProperty()
    unique_prop = StringProperty(unique_index=True)
    title = StringProperty(required=True, indexed=True)
    publication_date = DateProperty()
    number_of_words = IntegerProperty()
    embedding = ArrayProperty(FloatProperty())

    # Outgoing rels
    has_author = RelationshipTo("Author", "HAS_AUTHOR")
    has_description = RelationshipTo("Description", "HAS_DESCRIPTION")
    has_abstract = RelationshipTo("Abstract", "HAS_ABSTRACT")

    # Incoming rel
    approved_by = RelationshipFrom("Approval", "APPROVED")

    # Same-label rel
    cites = RelationshipTo("Document", "CITES")


class Author(StructuredNode):
    name = StringProperty(index=True)

    in_office = RelationshipTo("Office", "IN_OFFICE")


class Office(StructuredNode):
    location = PointProperty(unique_index=True, crs="cartesian")


class Approval(StructuredNode):
    approval_datetime = DateTimeProperty()
    approval_local_datetime = DateTimeFormatProperty()
    approved = BooleanProperty(default=False)

    approved_by = RelationshipTo("Author", "APPROVED_BY")


class Description(StructuredNode):
    uid = UniqueIdProperty()
    content = StringProperty()


class Abstract(StructuredNode):
    uid = UniqueIdProperty()
    content = StringProperty()


class AsyncNeighbour(AsyncStructuredNode):
    uid = UniqueIdProperty()
    name = StringProperty()

    has_async_neighbour = AsyncRelationshipTo("AsyncNeighbour", "HAS_ASYNC_NEIGHBOUR")
    has_other_async_neighbour = AsyncRelationshipTo(
        "OtherAsyncNeighbour", "HAS_OTHER_ASYNC_NEIGHBOUR"
    )


class OtherAsyncNeighbour(AsyncStructuredNode):
    uid = UniqueIdProperty()
    unique_prop = StringProperty(unique_index=True)
    order = IntegerProperty(required=True, indexed=True)
